# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/semantics.py

"""Semantic analysis of propositional-logic constructs."""

from typing import AbstractSet, Iterable, Iterator, Mapping, Sequence, Tuple
from itertools import *
from .syntax import *
from .proofs import *

#: A model for propositional-logic formulas, a mapping from variable names to truth values.
Model = Mapping[str, bool]

def is_model(model: Model) -> bool:
    """Checks if the given dictionary is a model over some set of variable
    names.

    Parameters:
        model: dictionary to check.

    Returns:
        ``True`` if the given dictionary is a model over some set of variable
        names, ``False`` otherwise.
    """
    for key in model:
        if not is_variable(key):
            return False
    return True

def variables(model: Model) -> AbstractSet[str]:
    """Finds all variable names over which the given model is defined.

    Parameters:
        model: model to check.

    Returns:
        A set of all variable names over which the given model is defined.
    """
    assert is_model(model)
    return model.keys()

def evaluate(formula: Formula, model: Model) -> bool:
    """Calculates the truth value of the given formula in the given model.

    Parameters:
        formula: formula to calculate the truth value of.
        model: model over (possibly a superset of) the variable names of the
            given formula, to calculate the truth value in.

    Returns:
        The truth value of the given formula in the given model.

    Examples:
        >>> evaluate(Formula.parse('~(p&q76)'), {'p': True, 'q76': False})
        True

        >>> evaluate(Formula.parse('~(p&q76)'), {'p': True, 'q76': True})
        False
    """
    assert is_model(model)
    assert formula.variables().issubset(variables(model))
    root = formula.root
    if is_constant(root):
        return formula.root == 'T'
    elif is_variable(root):
        return model[root]
    elif is_unary(root):
        return not evaluate(formula.first, model)
    elif is_binary(root):
        if root == '&':
            return evaluate(formula.first, model) and evaluate(formula.second, model)
        elif root == '|':
            return evaluate(formula.first, model) or evaluate(formula.second, model)
        elif root == '->':
            return not evaluate(formula.first, model) or evaluate(formula.second, model)
        elif root == '-&':
            return not (evaluate(formula.first, model) and evaluate(formula.second, model))
        elif root == '-|':
            return not (evaluate(formula.first, model) or evaluate(formula.second, model))
        elif root == '+':
            return evaluate(formula.first, model) != evaluate(formula.second, model)
        elif root == '<->':
            return evaluate(formula.first, model) == evaluate(formula.second, model)
    # Task 2.1

def all_models(variables: Sequence[str]) -> Iterable[Model]:
    """Calculates all possible models over the given variable names.

    Parameters:
        variables: variable names over which to calculate the models.

    Returns:
        An iterable over all possible models over the given variable names. The
        order of the models is lexicographic according to the order of the given
        variable names, where False precedes True.

    Examples:
        >>> list(all_models(['p', 'q']))
        [{'p': False, 'q': False}, {'p': False, 'q': True}, {'p': True, 'q': False}, {'p': True, 'q': True}]

        >>> list(all_models(['q', 'p']))
        [{'q': False, 'p': False}, {'q': False, 'p': True}, {'q': True, 'p': False}, {'q': True, 'p': True}]
    """
    return (({tuple(variables)[j]:i[j] for j in range(len(i))}) for i in product({False, True}, repeat=len(variables)))
    # Task 2.2

def truth_values(formula: Formula, models: Iterable[Model]) -> Iterable[bool]:
    """Calculates the truth value of the given formula in each of the given
    models.

    Parameters:
        formula: formula to calculate the truth value of.
        models: iterable over models to calculate the truth value in.

    Returns:
        An iterable over the respective truth values of the given formula in
        each of the given models, in the order of the given models.

    Examples:
        >>> list(truth_values(Formula.parse('~(p&q76)'), all_models(['p', 'q76'])))
        [True, True, True, False]
    """
    return (evaluate(formula, model) for model in models)
    # Task 2.3

def print_truth_table(formula: Formula) -> None:
    """Prints the truth table of the given formula, with variable-name columns
    sorted alphabetically.

    Parameters:
        formula: formula to print the truth table of.

    Examples:
        >>> print_truth_table(Formula.parse('~(p&q76)'))
        | p | q76 | ~(p&q76) |
        |---|-----|----------|
        | F | F   | T        |
        | F | T   | T        |
        | T | F   | T        |
        | T | T   | F        |
    """
    def bool_to_str(value):
        assert type(value) == bool
        return 'T' if value else 'F'
    variables = sorted(formula.variables())
    print('| ' + ' | '.join([i for i in variables]) + ' | ' + str(formula) + ' |')
    print('|-' + '-|-'.join(['-'*len(i) for i in variables]) + '-|-' + '-'*len(str(formula)) + '-|')
    for i in all_models(variables):
        print('| ' + ' | '.join([bool_to_str(i[j])+' '*(len(j)-1) for j in variables]) + ' | ' + bool_to_str(evaluate(formula,i)) + ' ' * len(str(formula)) + '|')
    # Task 2.4

def is_tautology(formula: Formula) -> bool:
    """Checks if the given formula is a tautology.

    Parameters:
        formula: formula to check.

    Returns:
        ``True`` if the given formula is a tautology, ``False`` otherwise.
    """
    return all(truth_values(formula, all_models(Formula.variables(formula))))
    # Task 2.5a

def is_contradiction(formula: Formula) -> bool:
    """Checks if the given formula is a contradiction.

    Parameters:
        formula: formula to check.

    Returns:
        ``True`` if the given formula is a contradiction, ``False`` otherwise.
    """
    return not any(truth_values(formula, all_models(Formula.variables(formula))))
    # Task 2.5b

def is_satisfiable(formula: Formula) -> bool:
    """Checks if the given formula is satisfiable.

    Parameters:
        formula: formula to check.

    Returns:
        ``True`` if the given formula is satisfiable, ``False`` otherwise.
    """
    return any(truth_values(formula, all_models(Formula.variables(formula))))
    # Task 2.5c

def _synthesize_for_model(model: Model) -> Formula:
    """Synthesizes a propositional formula in the form of a single conjunctive
    clause that evaluates to ``True`` in the given model, and to ``False`` in
    any other model over the same variable names.

    Parameters:
        model: model over a nonempty set of variable names, in which the
            synthesized formula is to hold.

    Returns:
        The synthesized formula.
    """
    assert is_model(model)
    assert len(model.keys()) > 0
    variables = list(model.keys())
    if not model[variables[0]]:
        formula = Formula('~', Formula(variables.pop(0)))
    else:
        formula = Formula(variables.pop(0))
    while variables:
        if not model[variables[0]]:
            second = Formula('~', Formula(variables.pop(0)))
        else:
            second = Formula(variables.pop(0))
        formula = Formula('&', formula, second)
    return formula

    # Task 2.6

def synthesize(variables: Sequence[str], values: Iterable[bool]) -> Formula:
    """Synthesizes a propositional formula in DNF over the given variable names,
    that has the specified truth table.

    Parameters:
        variables: nonempty set of variable names for the synthesized formula.
        values: iterable over truth values for the synthesized formula in every
            possible model over the given variable names, in the order returned
            by `all_models`\ ``(``\ `~synthesize.variables`\ ``)``.

    Returns:
        The synthesized formula.

    Examples:
        >>> formula = synthesize(['p', 'q'], [True, True, True, False])
        >>> for model in all_models(['p', 'q']):
        ...     evaluate(formula, model)
        True
        True
        True
        False
    """
    variables = list(variables)
    rows = zip(all_models(variables),values)
    disjuncts = [_synthesize_for_model(i[0]) for i in rows if i[1]]
    if disjuncts:
        formula = disjuncts.pop(0)
        while disjuncts:
            formula = Formula('|',formula,disjuncts.pop(0))
    else:
        var = Formula(variables.pop(0))
        formula = Formula('&',var, Formula('~',var))
        while variables:
            second = Formula(variables.pop(0))
            second = Formula('&',second, Formula('~',second))
            formula = Formula('|',formula, second)
    return formula
    # Task 2.7

def _synthesize_for_all_except_model(model: Model) -> Formula:
    """Synthesizes a propositional formula in the form of a single disjunctive
    clause that evaluates to ``False`` in the given model, and to ``True`` in
    any other model over the same variable names.

    Parameters:
        model: model over a nonempty set of variable names, in which the
            synthesized formula is to not hold.

    Returns:
        The synthesized formula.
    """
    assert is_model(model)
    assert len(model.keys()) > 0
    variables = list(model.keys())
    init = True
    for i in variables:
        if init:
            if model[i]:
                formula = Formula('~',Formula(i))
                init = False
            else:
                formula = Formula(i)
                init = False
        else:
            if model[i]:
                first = Formula('~',Formula(i))
            else:
                first = Formula(i)
            formula = Formula('|', formula, first)
    return formula
    # Optional Task 2.8

def synthesize_cnf(variables: Sequence[str], values: Iterable[bool]) -> Formula:
    """Synthesizes a propositional formula in CNF over the given variable names,
    that has the specified truth table.

    Parameters:
        variables: nonempty set of variable names for the synthesized formula.
        values: iterable over truth values for the synthesized formula in every
            possible model over the given variable names, in the order returned
            by `all_models`\ ``(``\ `~synthesize.variables`\ ``)``.

    Returns:
        The synthesized formula.

    Examples:
        >>> formula = synthesize_cnf(['p', 'q'], [True, True, True, False])
        >>> for model in all_models(['p', 'q']):
        ...     evaluate(formula, model)
        True
        True
        True
        False
    """
    assert len(variables) > 0
    variables = list(variables)
    rows = zip(all_models(variables),values)
    conjuncts = [_synthesize_for_all_except_model(i[0]) for i in rows if not i[1]]
    if conjuncts:
        formula = conjuncts.pop(0)
        while conjuncts:
            formula = Formula('&', formula, conjuncts.pop(0))
    else:
        var = Formula(variables.pop(0))
        formula = Formula('|',var, Formula('~',var))
        while variables:
            second = Formula(variables.pop(0))
            second = Formula('|',second, Formula('~',second))
            formula = Formula('&',formula, second)
    return formula
    # Optional Task 2.9

def combine_formulae(formulae: Iterable[Formula], operator: str) -> Formula:
    if len(formulae) == 0:
        return formulae
    elif len(formulae) == 1:
        return formulae[0]
    else:
        return Formula(operator, formulae[0], combine_formulae(formulae[1:], operator))
    # Personal task

def evaluate_inference(rule: InferenceRule, model: Model) -> bool:
    """Checks if the given inference rule holds in the given model.

    Parameters:
        rule: inference rule to check.
        model: model to check in.

    Returns:
        ``True`` if the given inference rule holds in the given model, ``False``
        otherwise.

    Examples:
        >>> evaluate_inference(InferenceRule([Formula('p')], Formula('q')),
        ...                    {'p': True, 'q': False})
        False

        >>> evaluate_inference(InferenceRule([Formula('p')], Formula('q')),
        ...                    {'p': False, 'q': False})
        True
    """
    assert is_model(model)
    if rule.assumptions:
        formula = Formula('->', combine_formulae(rule.assumptions, '&'), rule.conclusion)
    else:
        formula = rule.conclusion
    return evaluate(formula, model)
    # Task 4.2

def is_sound_inference(rule: InferenceRule) -> bool:
    """Checks if the given inference rule is sound, i.e., whether its
    conclusion is a semantically correct implication of its assumptions.

    Parameters:
        rule: inference rule to check.

    Returns:
        ``True`` if the given inference rule is sound, ``False`` otherwise.
    """
    return all((evaluate_inference(rule, model) for model in all_models(InferenceRule.variables(rule))))
    # Task 4.3



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/operators.py

"""Syntactic conversion of predicate formulas to use only specific sets of
operators."""
from typing import Mapping
from .syntax import ForbiddenVariableError, Term, Formula
from .proofs import Schema

def substitute_operators(self, substitution_map: Mapping[str, Formula]) -> Formula:
    """Substitutes in the current formula, each constant or operator `op`
    that is a key in `substitution_map` with the formula
    `substitution_map[op]` applied to its (zero or one or two) operands,
    where the first operand is used for every occurrence of ``'p'`` in the
    formula and the second for every occurrence of ``'q'``.

    Parameters:
        substitution_map: mapping defining the substitutions to be
            performed.

    Returns:
        The formula resulting from performing all substitutions. Only
        operator occurrences originating in the current formula are
        substituted (i.e., operator occurrences originating in one of the
        specified substitutions are not subjected to additional
        substitutions).

    Examples:
        >>> substitute_operators(Formula.parse('((R(a,b)&S(f(a)))->a=b)'),
        ...     {'&': Schema(Formula.parse('~(~P()|~Q())'), {'P','Q'})})
        (~(~R(a,b)|~S(f(a)))->a=b)
    """
    for operator in substitution_map:
        assert is_unary(operator) or is_binary(operator)
        assert substitution_map[operator].templates.issubset({'P', 'Q'})
    root = self.root

    substitution_map = {operator:substitution_map[operator] for operator in substitution_map if operator in Formula.operators(self)}
    if substitution_map:
        if is_quantifier(root):
            result = Formula(root, self.variable, substitute_operators(self.statement, substitution_map))

        if is_unary(root):
            if root in substitution_map:
                first = substitute_operators(self.first, substitution_map)
                inst_map = {'P': first}
                result = substitution_map[root].instantiate(inst_map)
            else:
                result = Formula(root, substitute_operators(self.first, substitution_map))

        elif is_binary(root):
            if root in substitution_map:
                first = substitute_operators(self.first, substitution_map)
                second = substitute_operators(self.second, substitution_map)
                inst_map = {'P': first, 'Q': second}
                result = substitution_map[root].instantiate(inst_map)
            else:
                result = Formula(root, substitute_operators(self.first, substitution_map), substitute_operators(self.second, substitution_map))
    else:
        result = self

    return result
    # Harris J. Bolus - Personal Task

def to_not_and_or(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '~', '&', and '|'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '~', '&', and
        '|'.
    """
    sub_map = {'->': Schema(Formula.parse('(~P()|Q())'), {'P', 'Q'}),
               '+': Schema(Formula.parse('((P()|Q())&~(P()&Q()))'), {'P', 'Q'}),
               '-&': Schema(Formula.parse('~(P()&Q())'), {'P', 'Q'}),
               '-|': Schema(Formula.parse('~(P()|Q())'), {'P', 'Q'}),
               '<->': Schema(Formula.parse('((~P()|Q())&(~Q()|P()))'), {'P', 'Q'})}
    return substitute_operators(formula, sub_map)
    # Task 3.5

def to_not_and(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '~' and '&'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '~' and '&'.
    """
    sub_map = {'|': Schema(Formula.parse('~(~P()&~Q())'), {'P', 'Q'}),
               '->': Schema(Formula.parse('~(P()&~Q())'), {'P', 'Q'}),
               '+': Schema(Formula.parse('(~(P()&Q())&~(~P()&~Q()))'), {'P', 'Q'}),
               '-&': Schema(Formula.parse('~(P()&Q())'), {'P', 'Q'}),
               '-|': Schema(Formula.parse('(~P()&~Q())'), {'P', 'Q'}),
               '<->': Schema(Formula.parse('(~(P()&~Q())&~(Q()&~P()))'), {'P', 'Q'})}
    return substitute_operators(formula, sub_map)
    # Task 3.6a

def to_nand(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '-&'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '-&'.
    """
    sub_map = {'~': Schema(Formula.parse('(P()-&P())'), {'P'}),
               '&': Schema(Formula.parse('((P()-&Q())-&(P()-&Q()))'), {'P', 'Q'}),
               '|': Schema(Formula.parse('((P()-&P())-&(Q()-&Q()))'), {'P', 'Q'}),
               '->': Schema(Formula.parse('(P()-&(Q()-&Q()))'), {'P', 'Q'}),
               '+': Schema(Formula.parse('((P()-&(P()-&Q()))-&(Q()-&(P()-&Q())))'), {'P', 'Q'}),
               '-|': Schema(Formula.parse('(((P()-&Q())-&(Q()-&Q()))-&((P()-&P())-&(Q()-&Q())))'), {'P', 'Q'}),
               '<->': Schema(Formula.parse('(((P()-&P())-&(Q()-&Q()))-&(P()-&Q()))'), {'P', 'Q'})}
    return substitute_operators(formula, sub_map)
    # Task 3.6b

def to_implies_not(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '->' and '~'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '->' and '~'.
    """
    sub_map = {'&': Schema(Formula.parse('~(P()->~Q())'), {'P', 'Q'}),
               '|': Schema(Formula.parse('(~P()->Q())'), {'P', 'Q'}),
               '+': Schema(Formula.parse('((P()->Q())->~(Q()->P()))'), {'P', 'Q'}),
               '-|': Schema(Formula.parse('~(~P()->Q())'), {'P', 'Q'}),
               '-&': Schema(Formula.parse('(P()->~Q())'), {'P', 'Q'}),
               '<->': Schema(Formula.parse('~((P()->Q())->~(Q()->P()))'), {'P', 'Q'})}
    return substitute_operators(formula, sub_map)
    # Task 3.6c

def to_not_or(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '~' and '|'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '~' and '|'.
    """
    sub_map = {'&': Schema(Formula.parse('~(~P()|~Q())'), {'P', 'Q'}),
               '->': Schema(Formula.parse('~(~P()|Q())'), {'P', 'Q'}),
               '+': Schema(Formula.parse('(~(~P()|Q())|~(~Q()|P()))'), {'P', 'Q'}),
               '-&': Schema(Formula.parse('(~P()|~Q())'), {'P', 'Q'}),
               '-|': Schema(Formula.parse('~(P()|Q())'), {'P', 'Q'}),
               '<->': Schema(Formula.parse('~(~(~P()|Q())|~(~Q()|P()))'), {'P', 'Q'})}
    return substitute_operators(formula, sub_map)
    # Personal task

def to_nor(formula: Formula) -> Formula:
    """Syntactically converts the given formula to an equivalent formula that
    contains no constants or operators beyond '-|'.

    Parameters:
        formula: formula to convert.

    Returns:
        A formula that has the same truth table as the given formula, but
        contains no constants or operators beyond '-|'.
    """
    sub_map = {'~': Schema(Formula.parse('(P()-|P())'), {'P'}),
               '&': Schema(Formula.parse('((P()-|P())-|(Q()-|Q()))'), {'P', 'Q'}),
               '|': Schema(Formula.parse('((P()-|Q())-|(P()-|Q()))'), {'P', 'Q'}),
               '->': Schema(Formula.parse('(((P()-|P())-|Q())-|((P()-|P())-|Q()))'), {'P', 'Q'}),
               '+': Schema(Formula.parse('((P()-|Q())-|((P()-|P())-|(Q()-|Q())))'), {'P', 'Q'}),
               '-&': Schema(Formula.parse('(((P()-|P())-|(Q()-|Q()))-|((P()-|P())-|(Q()-|Q())))'), {'P', 'Q'}),
               '<->': Schema(Formula.parse('((P()-|(P()-|Q()))-|(Q()-|(P()-|Q())))'), {'P', 'Q'})}
    return substitute_operators(formula, sub_map)
    # Personal task

def frege(formula: Formula) -> None:
    """Computes a representation of the current formula in Frege's notation.

    Returns:
        A Fregean notation representation of the current formula.
    """

    def _frege_helper(self: Formula) -> str:
        """Computes a representation of the current formula, which contains only
       the operators `~` and `->`, in Frege's notation, lacking a judgement stroke.

        Returns:
            The Fregean notation representation of the current formula, lacking
            a judgement stroke.
        """

        root = self.root
        if is_relation(root) or is_equality(root):
            return ['─ '+str(self)]

        elif is_unary(root):
            first = _frege_helper(self.first)
            if len(first) > 1:
                for i in range(1, len(first)):
                    first[i] = '  '+first[i]
            first[0] = '─┬'+first[0]
            return first

        elif is_binary(root):
            first = _frege_helper(self.first)
            second = _frege_helper(self.second)
            diff = len(first[0].split(' ')[0])-len(second[0].split(' ')[0])
            if diff > 0:
                second[0] = '─'*diff+second[0]
            second[0] = '─┬'+second[0]
            if len(second) > 1:
                for i in range(1, len(second)):
                    if diff > 0:
                        second[i] = ' '*diff+second[i]
                    second[i] = ' │'+second[i]

            if diff < 0:
                first[0] = '─'*(diff*-1)+first[0]
            first[0] = ' └'+first[0]
            if len(first) > 1:
                for i in range(1, len(first)):
                    if diff < 0:
                        first[i] = ' '*(diff*-1)+first[i]
                    first[i] = '  '+first[i]
            return [*second, *first]

        else:
            assert is_quantifier(root)
            variable = self.variable
            statement = _frege_helper(self.statement)

            if root == 'A':
                statement[0] = f'─{variable}─{statement[0]}'
            elif root == 'E':
                statement[0] = f'┬{variable}┬{statement[0]}'

            for i in range(1, len(statement)):
                statement[i] = (len(variable)+2)*' '+statement[i]
            return statement
    string = _frege_helper(to_implies_not(formula))
    print('├─'+'\n  '.join(string))



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/semantics.py

"""Semantic analysis of predicate-logic expressions."""

from typing import AbstractSet, FrozenSet, Generic, Mapping, Tuple, TypeVar, Set, Mapping
from datetime import datetime
from mltp.logic_utils import frozen, frozendict
from .syntax import ForbiddenVariableError, Term, Formula, is_constant, is_variable, is_function, is_equality, is_relation, is_unary, is_binary, is_quantifier
from mltp.propositions.semantics import print_truth_table as print_propositional_truth_table

#: A generic type for a universe element in a model.
T = TypeVar('T')

@frozen
class Model(Generic[T]):
    """An immutable model for predicate-logic constructs.

    Attributes:
        universe (`~typing.FrozenSet`\\[`T`]): the set of elements to which
            terms can be evaluated and over which quantifications are defined.
        constant_interpretations (`~typing.Mapping`\\[`str`, `T`]): mapping from
            each constant name to the universe element to which it evaluates.
        relation_arities (`~typing.Mapping`\\[`str`, `int`]): mapping from
            each relation name to its arity, or to ``-1`` if the relation is the
            empty relation.
        relation_interpretations (`~typing.Mapping`\\[`str`, `~typing.AbstractSet`\\[`~typing.Tuple`\\[`T`, ...]]]):
            mapping from each n-ary relation name to argument n-tuples (of
            universe elements) for which the relation is true.
        function_arities (`~typing.Mapping`\\[`str`, `int`]): mapping from
            each function name to its arity.
        function_interpretations (`~typing.Mapping`\\[`str`, `~typing.Mapping`\\[`~typing.Tuple`\\[`T`, ...], `T`]]):
            mapping from each n-ary function name to the mapping from each
            argument n-tuple (of universe elements) to the universe element that
            the function outputs given these arguments.
    """
    universe: FrozenSet[T]
    constant_interpretations: Mapping[str, T]
    relation_arities: Mapping[str, int]
    relation_interpretations: Mapping[str, AbstractSet[Tuple[T, ...]]]
    function_arities: Mapping[str, int]
    function_interpretations: Mapping[str, Mapping[Tuple[T, ...], T]]

    def __init__(self, universe: AbstractSet[T],
                 constant_interpretations: Mapping[str, T],
                 relation_interpretations:
                 Mapping[str, AbstractSet[Tuple[T, ...]]],
                 function_interpretations:
                 Mapping[str, Mapping[Tuple[T, ...], T]] = frozendict()):
        """Initializes a `Model` from its universe and constant, relation, and
        function name interpretations.

        Parameters:
            universe: the set of elements to which terms are to be evaluated
                and over which quantifications are to be defined.
            constant_interpretations: mapping from each constant name to a
                universe element to which it is to be evaluated.
            relation_interpretations: mapping from each relation name that is to
                be the name of an n-ary relation, to the argument n-tuples (of
                universe elements) for which the relation is to be true.
            function_interpretations: mapping from each function name that is to
                be the name of an n-ary function, to a mapping from each
                argument n-tuple (of universe elements) to a universe element
                that the function is to output given these arguments.
        """
        for constant in constant_interpretations:
            assert is_constant(constant), print(f'{constant} is not a valid constant name')
            assert constant_interpretations[constant] in universe, print(f'the constant {constant_interpretations[constant]} is not in the domain')
        relation_arities = {}
        for relation in relation_interpretations:
            assert is_relation(relation), print(f'{relation} is not a valid relation')
            relation_interpretation = relation_interpretations[relation]
            if len(relation_interpretation) == 0:
                arity = -1 # any
            else:
                some_arguments = next(iter(relation_interpretation))
                arity = len(some_arguments)
                for arguments in relation_interpretation:
                    assert len(arguments) == arity
                    for argument in arguments:
                        assert argument in universe, print(f'argument {argument} not in universe, type: {type(argument)}')
            relation_arities[relation] = arity
        function_arities = {}
        for function in function_interpretations:
            assert is_function(function), print(f'{function} is not a valid function. Function names must begin with a lowercase letter f-t and must be alphanumeric.')
            function_interpretation = function_interpretations[function]
            assert len(function_interpretation) > 0
            some_argument = next(iter(function_interpretation))
            arity = len(some_argument)
            assert arity > 0
            assert len(function_interpretation) == len(universe)**arity, print(f'\nFunction interpretations are: {function_interpretation}, \n universe is {universe}\n')
            for arguments in function_interpretation:
                assert len(arguments) == arity
                for argument in arguments:
                    assert argument in universe
                assert function_interpretation[arguments] in universe
            function_arities[function] = arity

        self.universe = frozenset(universe)
        self.constant_interpretations = frozendict(constant_interpretations)
        self.relation_arities = frozendict(relation_arities)
        self.relation_interpretations = \
            frozendict({relation: frozenset(relation_interpretations[relation])
                        for relation in relation_interpretations})
        self.function_arities = frozendict(function_arities)
        self.function_interpretations = \
            frozendict({function: frozendict(function_interpretations[function])
                        for function in function_interpretations})
        self.max_cache_size = 100
        self.formula_cache = {}


    def __repr__(self) -> str:
        """Computes a string representation of the current model.

        Returns:
            A string representation of the current model.
        """
        return 'Universe=' + str(self.universe) + \
               '; Constant Interpretations=' + \
               str(self.constant_interpretations) + \
               '; Relation Interpretations=' + \
               str(self.relation_interpretations) + \
               ('; Function Interpretations=' + \
                str(self.function_interpretations)
                if len(self.function_interpretations) > 0 else '')

    def evaluate_term(self, term: Term,
                      assignment: Mapping[str, T] = frozendict()) -> T:
        """Calculates the value of the given term in the current model under the
        given assignment of values to variable names.

        Parameters:
            term: term to calculate the value of, for the constant and function
                names of which the current model has interpretations.
            assignment: mapping from each variable name in the given term to a
                universe element to which it is to be evaluated.

        Returns:
            The value (in the universe of the current model) of the given
            term in the current model under the given assignment of values to
            variable names.
        """
        assert term.constants().issubset(self.constant_interpretations.keys())
        assert term.variables().issubset(assignment.keys()), print(f'variables in term are {term.variables()}, assignment includes {assignment.keys()}')
        for function,arity in term.functions():
            assert function in self.function_interpretations and self.function_arities[function] == arity

        root = term.root
        if is_constant(root):
            result = self.constant_interpretations[root]
        elif is_variable(root):
            result = assignment[root]
        elif is_function(root):
            result = self.function_interpretations[root][tuple(self.evaluate_term(argument, assignment) for argument in term.arguments)]

        return result
        # Task 7.7

    def evaluate_formula(self, formula: Formula,
                         assignment: Mapping[str, T] = frozendict()) -> bool:
        assert formula.constants().issubset(self.constant_interpretations.keys())
        assert formula.free_variables().issubset(assignment.keys()), \
            f'in the formula {formula}, formula.free_variables() is {formula.free_variables()}, assignment.keys() is {assignment.keys()}'
        for function, arity in formula.functions():
            assert function in self.function_interpretations and \
                   self.function_arities[function] == arity
        for relation, arity in formula.relations():
            assert relation in self.relation_interpretations and \
                   self.relation_arities[relation] in {-1, arity}

        key = (formula, frozenset(assignment.items()))
        cache = self.formula_cache
        if key in cache:
            return cache[key]
        if len(cache) > self.max_cache_size:
            cache.pop(next(iter(cache)))

        root = formula.root
        if is_equality(root):
            result = self.evaluate_term(formula.arguments[0], assignment) is self.evaluate_term(formula.arguments[1], assignment)
        elif is_relation(root):
            result = tuple(self.evaluate_term(argument, assignment) for argument in formula.arguments) in self.relation_interpretations[root]
        elif is_unary(root):
            result = not self.evaluate_formula(formula.first, assignment)
        elif is_binary(root):
            if root == '&':
                result = self.evaluate_formula(formula.first, assignment) and self.evaluate_formula(formula.second, assignment)
            elif root == '|':
                result = self.evaluate_formula(formula.first, assignment) or self.evaluate_formula(formula.second, assignment)
            elif root == '->':
                result = not self.evaluate_formula(formula.first, assignment) or self.evaluate_formula(formula.second, assignment)
            elif root == '-&':
                result = not (self.evaluate_formula(formula.first, assignment) and self.evaluate_formula(formula.second, assignment))
            elif root == '-|':
                result = not (self.evaluate_formula(formula.first, assignment) or self.evaluate_formula(formula.second, assignment))
            elif root == '+':
                result = self.evaluate_formula(formula.first, assignment) != self.evaluate_formula(formula.second, assignment)
            elif root == '<->':
                result = self.evaluate_formula(formula.first, assignment) == self.evaluate_formula(formula.second, assignment)

        elif is_quantifier(root):
            variable = formula.variable
            new_assignments = ({variable: value} for value in self.universe)
            truth_values = (self.evaluate_formula(formula.statement, {**assignment, **new_assignment}) for new_assignment in new_assignments)
            if root == 'A':
                result = all(truth_values)
            else:
                result = any(truth_values)

        cache[key] = result
        return result
        # Task 7.8

    def find_assignments(self, formula: Formula, assignment: Mapping[str, T] = frozendict(), value: bool = True) -> Set[Formula]:
        """Returns a generator object of assignments for which the formula is true in the model"""

        assert is_quantifier(formula.root), print(f'{formula} is not quantified')
        variable = formula.variable
        new_assignments = ({variable:value} for value in self.universe)
        truth_values = (({**assignment, **new_assignment}, self.evaluate_formula(formula.statement, {**assignment, **new_assignment})) for new_assignment in new_assignments)
        if value:
            return [i[0] for i in truth_values if i[1]]
        else:
            return [i[0] for i in truth_values if not i[1]] #only gets the first assignment
        # Personal task

    def is_model_of(self, formulas: AbstractSet[Formula]) -> bool:
        """Checks if the current model is a model of the given formulas.

        Parameters:
            formulas: formulas to check, for the constant, function, and
                relation names of which the current model has interpretations.

        Returns:
            ``True`` if each of the given formulas evaluates to true in the
            current model under any assignment of elements from the universe of
            the current model to the free occurrences of variable names in that
            formula, ``False`` otherwise.
        """
        for formula in formulas:
            assert formula.constants().issubset(
                self.constant_interpretations.keys())
            for function,arity in formula.functions():
                assert function in self.function_interpretations and \
                       self.function_arities[function] == arity
            for relation,arity in formula.relations():
                assert relation in self.relation_interpretations and \
                       self.relation_arities[relation] in {-1, arity}
        new_formulas = set()
        for formula in formulas:
            for free_variable in formula.free_variables():
                formula = Formula('A', free_variable, formula)
            new_formulas.add(formula)
        return all(self.evaluate_formula(formula) for formula in new_formulas)
        # Task 7.9

# minimum = 0
# maximum = 3
# universe = set(str(i) for i in range(minimum,maximum+1)) | {'undefined'}
# relation_interpretations = {'L':set(),'GorE':set()}
# function_interpretations = {'plus':dict(), 'minus':dict(), 'times':dict(), 'fdividedby': dict()}
# for i in universe:
#     for j in universe:
#         if i == 'undefined' or j == 'undefined':
#             function_interpretations['plus'][(i,j)] = 'undefined'
#             function_interpretations['minus'][(i,j)] = 'undefined'
#             function_interpretations['times'][(i,j)] = 'undefined'
#             function_interpretations['fdividedby'][(i,j)] = 'undefined'
#             continue
#         if int(i)+int(j) <= maximum:
#             function_interpretations['plus'][(i,j)] = str(int(i)+int(j))
#         else:
#             function_interpretations['plus'][(i,j)] = 'undefined'

#         if int(i)-int(j) >= minimum:
#             function_interpretations['minus'][(i,j)] = str(int(i)-int(j))
#         else:
#             function_interpretations['minus'][(i,j)] = 'undefined'

#         if int(i)*int(j) <= maximum:
#             function_interpretations['times'][(i,j)] = str(int(i)*int(j))
#         else:
#             function_interpretations['times'][(i,j)] = 'undefined'

#         if j != '0' and int(i) % int(j) == 0:
#             function_interpretations['fdividedby'][(i,j)] = str(int(int(i)/int(j)))
#         else:
#             function_interpretations['fdividedby'][(i,j)] = 'undefined'

#         if i >= j:
#             relation_interpretations['GorE'].add((i,j))
#         else:
#             relation_interpretations['L'].add((i,j))

# model = Model(
#                 universe,                                                                       #universe
#                 {(i if is_constant(i) else make_constant(i,universe)):i for i in universe},     #constant interpretations
#                 relation_interpretations,                                                       #relation interpretations
#                 function_interpretations                                                        #function interpretations
#                 )

# for i in range(1,6):
#     formula = make_formula_of_size(i)

#     startTime = datetime.now()
#     for j in range(10):
#         model.evaluate_formula(formula)
#     time1 = datetime.now() - startTime

#     startTime = datetime.now()
#     for j in range(10):
#         model.evaluate_formula(formula)
#     time2 = datetime.now() - startTime

#     print(f'at length {i}, original is {time1} and numpy fromiter is {time2}')




# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: propositions/proofs.py

"""Proofs by deduction in Propositional Logic."""
from __future__ import annotations
from typing import AbstractSet, FrozenSet, List, Mapping, Optional, Sequence, Set, Tuple, Union
from mltp.logic_utils import frozen, memoized_parameterless_method
from .syntax import *

#: A mapping from variable names to formulas.
SpecializationMap = Mapping[str, Formula]

@frozen
class InferenceRule:
    """An immutable inference rule in Propositional Logic, comprised of zero
    or more assumed propositional formulas, and a conclusion propositional
    formula.

    Attributes:
        assumptions (`~typing.Tuple`\\[`~propositions.syntax.Formula`, ...]):
            the assumptions of the rule.
        conclusion (`~propositions.syntax.Formula`): the conclusion of the rule.
    """
    assumptions: Tuple[Formula, ...]
    conclusion: Formula

    def __init__(self, assumptions: Sequence[Formula], conclusion: Formula):
        """Initializes an `InferenceRule` from its assumptions and conclusion.

        Parameters:
            assumptions: the assumptions for the rule.
            conclusion: the conclusion for the rule.
        """
        self.assumptions = tuple(assumptions)
        self.conclusion = conclusion

    @memoized_parameterless_method
    def __repr__(self) -> str:
        """Computes a string representation of the current inference rule.

        Returns:
            A string representation of the current inference rule.
        """
        return str([str(assumption) for assumption in self.assumptions]) + ' ==> ' + "'" + str(self.conclusion) + "'"

    def __eq__(self, other: object) -> bool:
        """Compares the current inference rule with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is an `InferenceRule` object that
            equals the current inference rule, ``False`` otherwise.
        """
        return isinstance(other, InferenceRule) and \
               self.assumptions == other.assumptions and \
               self.conclusion == other.conclusion

    def __ne__(self, other: object) -> bool:
        """Compares the current inference rule with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is not an `InferenceRule` object or
            does not does not equal the current inference rule, ``False``
            otherwise.
        """
        return not self == other

    def __hash__(self) -> int:
        return hash(str(self))

    def variables(self) -> Set[str]:
        """Finds all variable names in the current inference rule.

        Returns:
            A set of all variable names used in the assumptions and in the
            conclusion of the current inference rule.
        """
        variable_list = set()
        for assumption in self.assumptions:
            variable_list = variable_list | Formula.variables(assumption)
        return variable_list | Formula.variables(self.conclusion)
        # Task 4.1

    def specialize(self, specialization_map: SpecializationMap) -> \
            InferenceRule:
        """Specializes the current inference rule by simultaneously substituting
        each variable name `v` that is a key in `specialization_map` with the
        formula `specialization_map[v]`.

        Parameters:
            specialization_map: mapping defining the specialization to be
                performed.

        Returns:
            The resulting inference rule.
        """
        for variable in specialization_map:
            assert is_variable(variable), print(f'{variable} is not a variable')
            if isinstance(specialization_map[variable], str):
                specialization_map[variable] = Formula.parse(specialization_map[variable])

        specialized_assumptions = (Formula.substitute_variables(assumption, specialization_map) for assumption in self.assumptions)
        specialized_conclusion = Formula.substitute_variables(self.conclusion, specialization_map)
        return InferenceRule(specialized_assumptions, specialized_conclusion)
        # Task 4.4

    @staticmethod
    def _merge_specialization_maps(
            specialization_map1: Union[SpecializationMap, None],
            specialization_map2: Union[SpecializationMap, None]) -> \
            Union[SpecializationMap, None]:
        """Merges the given specialization maps while checking their
        consistency.

        Parameters:
            specialization_map1: first mapping to merge, or ``None``.
            specialization_map2: second mapping to merge, or ``None``.

        Returns:
            A single mapping containing all (key, value) pairs that appear in
            either of the given maps, or ``None`` if one of the given maps is
            ``None`` or if some key appears in both given maps but with
            different values.
        """
        if specialization_map1 is not None:
            for variable in specialization_map1:
                assert is_variable(variable)
        else:
            return None
        if specialization_map2 is not None:
            for variable in specialization_map2:
                assert is_variable(variable)
        else:
            return None

        map1 = dict(specialization_map1)

        intersection = set(specialization_map1.keys()) & set(specialization_map2.keys())
        for key in intersection:
            if specialization_map1[key] != specialization_map2[key]:
                return None

        difference = set(specialization_map2.keys()) - set(specialization_map1.keys())
        for key in difference:
            map1[key] = specialization_map2[key]

        return map1
        # Task 4.5a

    @staticmethod
    def _formula_specialization_map(general: Formula, specialization: Formula) \
            -> Union[SpecializationMap, None]:
        """Computes the minimal specialization map by which the given formula
        specializes to the given specialization.

        Parameters:
            general: non-specialized formula for which to compute the map.
            specialization: specialization for which to compute the map.

        Returns:
            The computed specialization map, or ``None`` if `specialization` is
            in fact not a specialization of `general`.
        """
        #1. Break them down by tree structure until one operand is a variable, then close that branch.
        map1 = dict()
        if is_variable(general.root):
            map1 = {general.root: specialization}
            return map1

        elif is_constant(general.root):
            if general.root == specialization.root:
                return {}
            else:
                return None

        if general.root == specialization.root:
            if is_unary(general.root):
                map2 = InferenceRule._formula_specialization_map(general.first, specialization.first)
                if map2 != None:
                    map1 = InferenceRule._merge_specialization_maps(map1, map2)
                    return map1
                else:
                    return None

            elif is_binary(general.root):
                map2 = InferenceRule._formula_specialization_map(general.first, specialization.first)
                map3 = InferenceRule._formula_specialization_map(general.second, specialization.second)

                if map2 != None and map3 != None:
                    map1 = InferenceRule._merge_specialization_maps(map1, map2)
                    map1 = InferenceRule._merge_specialization_maps(map1, map3)
                    return map1

                else:
                    return None
        else:
            return None
        # Task 4.5b

    def specialization_map(self, specialization: InferenceRule) -> \
            Union[SpecializationMap, None]:
        """Computes the minimal specialization map by which the current
        inference rule specializes to the given specialization.

        Parameters:
            specialization: specialization for which to compute the map.

        Returns:
            The computed specialization map, or ``None`` if `specialization` is
            in fact not a specialization of the current rule.
        """
        map1 = dict()
        if len(self.assumptions) == len(specialization.assumptions):
            for i,j in zip(self.assumptions, specialization.assumptions):
                map1 = InferenceRule._merge_specialization_maps(map1, InferenceRule._formula_specialization_map(i,j))
        else:
            return None
        map1 = InferenceRule._merge_specialization_maps(map1, InferenceRule._formula_specialization_map(self.conclusion,specialization.conclusion))
        return map1 if map1 else None
        # Task 4.5c

    def is_specialization_of(self, general: InferenceRule) -> bool:
        """Checks if the current inference rule is a specialization of the given
        inference rule.

        Parameters:
            general: non-specialized inference rule to check.

        Returns:
            ``True`` if the current inference rule is a specialization of
            `general`, ``False`` otherwise.
        """
        return general.specialization_map(self) is not None

@frozen
class Proof:
    """An immutable deductive proof in Propositional Logic, comprised of a
    statement in the form of an inference rule, a set of inference rules that
    may be used in the proof, and a list of lines that prove the statement via
    these inference rules.

    Attributes:
        statement (`InferenceRule`): the statement proven by the proof.
        rules (`~typing.AbstractSet`\\[`InferenceRule`]): the allowed rules of
            the proof.
        lines (`~typing.Tuple`\\[`Line`]): the lines of the proof.
    """
    statement: InferenceRule
    rules: FrozenSet[InferenceRule]
    lines: Tuple[Proof.Line, ...]

    def __init__(self, statement: InferenceRule,
                 rules: AbstractSet[InferenceRule],
                 lines: Sequence[Proof.Line]):
        """Initializes a `Proof` from its statement, allowed inference rules,
        and lines.

        Parameters:
            statement: the statement to be proven by the proof.
            rules: the allowed rules for the proof.
            lines: the lines for the proof.
        """
        self.statement = statement
        self.rules = frozenset(rules)
        self.lines = tuple(lines)

    @frozen
    class Line:
        """An immutable line in a deductive proof, comprised of a formula that
        is justified either as an assumption of the proof, or as the conclusion
        of a specialization of an allowed inference rule of the proof, the
        assumptions of which are justified by previous lines in the proof.

        Attributes:
            formula (`~propositions.syntax.Formula`): the formula justified by
                the line.
            rule (`~typing.Optional`\\[`InferenceRule`]): the inference rule,
                out of those allowed in the proof, that has a specialization
                that concludes the formula; or ``None`` if the formula is
                justified as an assumption of the proof.
            assumptions (`~typing.Optional`\\[`~typing.Tuple`\\[`int`]): tuple
                of zero or more numbers of previous lines in the proof whose
                formulas are the respective assumptions of the specialization of
                the rule that concludes the formula, if the formula is not
                justified as an assumption of the proof.
        """
        formula: Formula
        rule: Optional[InferenceRule]
        assumptions: Optional[Tuple[int, ...]]

        def __init__(self, formula: Formula,
                     rule: Optional[InferenceRule] = None,
                     assumptions: Optional[Sequence[int]] = None):
            """Initializes a `~Proof.Line` from its formula, and optionally its
            rule and numbers of justifying previous lines.

            Parameters:
                formula: the formula to be justified by the line.
                rule: the inference rule, out of those allowed in the proof,
                    that has a specialization that concludes the formula; or
                    ``None`` if the formula is to be justified as an assumption
                    of the proof.
                assumptions: numbers of previous lines in the proof whose
                    formulas are the respective assumptions of the
                    specialization of the rule that concludes the formula, or
                    ``None`` if the formula is to be justified as an assumption
                    of the proof.
            """
            assert (rule is None and assumptions is None) or \
                   (rule is not None and assumptions is not None)
            self.formula = formula
            self.rule = rule
            if assumptions is not None:
                self.assumptions = tuple(assumptions)

        def __repr__(self) -> str:
            """Computes a string representation of the current line.

            Returns:
                A string representation of the current line.
            """
            if self.rule is None:
                return str(self.formula)
            else:
                r = str(self.formula) + '    (Inference Rule ' + str(self.rule)
                if len(self.assumptions) == 1:
                    r += ' on line ' + str(self.assumptions[0])
                elif len(self.assumptions) > 1:
                    r += ' on lines ' + ','.join(map(str, self.assumptions))
                r += ')'
                return r

        def is_assumption(self) -> bool:
            """Checks if the current proof line is justified as an assumption of
            the proof.

            Returns:
                ``True`` if the current proof line is justified as an assumption
                of the proof, ``False`` otherwise.
            """
            return self.rule is None

    def __repr__(self) -> str:
        """Computes a string representation of the current proof.

        Returns:
            A string representation of the current proof.
        """
        r = 'Proof of ' + str(self.statement) + ' via inference rules:\n'
        for rule in self.rules:
            r += '  ' + str(rule) + '\n'
        r += 'Lines:\n'
        for i in range(len(self.lines)):
            r += ('%3d) ' % i) + str(self.lines[i]) + '\n'
        r += "QED\n"
        return r

    def rule_for_line(self, line_number: int) -> Union[InferenceRule, None]:
        """Computes the inference rule whose conclusion is the formula justified
        by the specified line, and whose assumptions are the formulas justified
        by the lines specified as the assumptions of that line.

        Parameters:
            line_number: number of the line according to which to compute the
                inference rule.

        Returns:
            The computed inference rule, with assumptions ordered in the order
            of their numbers in the specified line, or ``None`` if the specified
            line is justified as an assumption.
        """
        assert line_number < len(self.lines)
        if Proof.Line.is_assumption(self.lines[line_number]):
            return None
        else:
            assumptions = list()
            for assumption in self.lines[line_number].assumptions:
                assumptions.append(self.lines[assumption].formula)
            conclusion = self.lines[line_number].formula
            return InferenceRule(assumptions, conclusion)
        # Task 4.6a

    def is_line_valid(self, line_number: int) -> bool:
        """Checks if the specified line validly follows from its justifications.

        Parameters:
            line_number: number of the line to check.

        Returns:
            If the specified line is justified as an assumption, then ``True``
            if the formula justified by this line is an assumption of the
            current proof, ``False`` otherwise. Otherwise (i.e., if the
            specified line is justified as a conclusion of an inference rule),
            ``True`` if the rule specified for that line is one of the allowed
            inference rules in the current proof, and it has a specialization
            that satisfies all of the following:

            1. The conclusion of that specialization is the formula justified by
               that line.
            2. The assumptions of this specialization are the formulas justified
               by the lines that are specified as the assumptions of that line
               (in the order of their numbers in that line), all of which must
               be previous lines.
        """
        assert line_number < len(self.lines)
        line = self.lines[line_number]
        if Proof.Line.is_assumption(line):
            return any(str(line) == str(assumption) for assumption in self.statement.assumptions)

        else:
            rule_was_given = any(line.rule == rule for rule in self.rules)
            rule_matches_line = Proof.rule_for_line(self, line_number).is_specialization_of(line.rule) or Proof.rule_for_line(self, line_number) == line.rule
            assumptions_are_prior = all(assumption < line_number for assumption in line.assumptions)
            assumptions_match_line = all(i == j for i,j in zip([self.lines[k].formula for k in line.assumptions], Proof.rule_for_line(self, line_number).assumptions))
            if not rule_was_given:
                print(f'rule was not given: {line}')
            if not rule_matches_line:
                print(f'rule does not match formula: {line}')
            if not assumptions_are_prior:
                print(f'assumptions are not found on prior lines: {line}')
            if not assumptions_match_line:
                print(f'assumptions do not match formula: {line}')
            return rule_was_given and rule_matches_line and assumptions_are_prior and assumptions_match_line
        # Task 4.6b

    def is_valid(self) -> bool:
        """Checks if the current proof is a valid proof of its claimed statement
        via its inference rules.

        Returns:
            ``True`` if the current proof is a valid proof of its claimed
            statement via its inference rules, ``False`` otherwise.
        """
        if self.lines:
            each_line_is_valid = all(self.is_line_valid(line) for line in range(len(self.lines)))
            if not each_line_is_valid:
                print([(str(line)+': '+str(self.is_line_valid(line))) for line in range(len(self.lines)) if not self.is_line_valid(line)])
            conclusion_matches = self.statement.conclusion == self.lines[-1].formula
            if not conclusion_matches:
                print('statement ' + str(self.statement.conclusion) + ' != conclusion ' + str(self.lines[-1].formula))
            return each_line_is_valid and conclusion_matches
        return False
        # Task 4.6c

    def cited_by(self, line_number) -> list:
        """Returns a list of lines cited by the given line"""

        line = self.lines[line_number]
        if not line.is_assumption():
            return line.assumptions
        # personal task

    def uncite_duplicate_lines(self) -> list:
        """Adjusts the assumptions of any line that cites a previous line with a duplicate
       so that its assumptions cite the first occurence of that formula."""

        formulae = [line.formula for line in self.lines]
        duplicates = []
        first_occurrences = {}
        for line_number in range(len(formulae)):
            formula = formulae[line_number]
            if formula in formulae[:line_number]:
                duplicates.append(line_number)
                if formula not in first_occurrences:
                    first_occurrences[formula] = line_number

        replacement_dict = {}
        for line_number in duplicates:
            replacement_dict[line_number] = first_occurrences[self.lines[line_number].formula]

        adjusted = []
        for line in self.lines:
            if line.is_assumption():
                adjusted.append(line)
            else:
                adjusted.append(Proof.Line(line.formula, line.rule, [replacement_dict[assumption] if assumption in replacement_dict else assumption for assumption in line.assumptions]))
        return Proof(self.statement, self.rules, adjusted)
        # personal task

    def uncited_lines(self) -> list:
        """Returns a list of all lines that are not cited by the conclusion, or
        not cited by any line that is cited by the conclusion."""
        #add conclusion
        cited = {len(self.lines)-1}
        temp = list(cited)
        #starting with conclusion, add cited lines to temp and cited. Follow up by finding lines cited by temp, and reset temp.
        while temp:
            current = temp.pop()
            new = self.cited_by(current)
            if new:
                for i in new:
                    temp.insert(0,i)
                    cited.add(i)
        return set(range(len(self.lines)))-cited
        # personal task

    def remove_uncited_lines(self) -> Proof:
        """Returns a new Proof of the same conclusion, with unncessary lines removed."""
        lines_to_remove = self.uncited_lines()
        lines_to_keep = sorted(set(i for i in range(len(self.lines))) - set(lines_to_remove))

        line_dict = {lines_to_keep[i]:i for i in range(len(lines_to_keep))}

        new_lines = []
        for line in range(len(self.lines)):
            if line not in lines_to_remove:
                new_lines.append(self.lines[line])
                line = self.lines[line]

        renumbered_lines = []
        for line in new_lines:
            if line.is_assumption():
                renumbered_lines.append(line)
            else:
                renumbered_lines.append(Proof.Line(line.formula, line.rule, [line_dict[assumption] if assumption in line_dict else assumption for assumption in line.assumptions]))
        return Proof(self.statement, self.rules, renumbered_lines)
        # personal task

    def clean(self) -> Proof:
        """Returns a proof with no duplicated or unnecessary lines."""
        proof = self.uncite_duplicate_lines()
        proof = self.remove_uncited_lines()
        print(f'Removed {len(self.lines)-len(proof.lines)} lines')
        assert self.is_valid() == proof.is_valid()
        return proof
        # personal task

def prove_specialization(proof: Proof, specialization: InferenceRule) -> Proof:
    """Converts the given proof of an inference rule to a proof of the given
    specialization of that inference rule.

    Parameters:
        proof: valid proof to convert.
        specialization: specialization of the rule proven by the given proof.

    Returns:
        A valid proof of the given specialization via the same inference rules
        as the given proof.
    """
    assert proof.is_valid()
    assert specialization.is_specialization_of(proof.statement)

    sub_map = proof.statement.specialization_map(specialization)
    specialized_statement = proof.statement.specialize(sub_map)
    specialized_lines = [Proof.Line(line.formula.substitute_variables(sub_map), line.rule, line.assumptions) if not Proof.Line.is_assumption(line) else Proof.Line(line.formula.substitute_variables(sub_map)) for line in proof.lines ]

    proof1 = Proof(specialized_statement, proof.rules, specialized_lines)

    return proof1
    # Task 5.1

def _inline_proof_once(main_proof: Proof, line_number: int,
                       lemma_proof: Proof) -> Proof:
    """Inlines the given proof of a "lemma" inference rule into the given proof
    that uses that "lemma" rule, eliminating the usage of (a specialization of)
    that "lemma" rule in the specified line in the latter proof.

    Parameters:
        main_proof: valid proof to inline into.
        line_number: number of the line in `main_proof` that should be replaced.
        lemma_proof: valid proof of the inference rule of the specified line (an
            allowed inference rule of `main_proof`).

    Returns:
        A valid proof obtained by replacing the specified line in `main_proof`
        with a full (specialized) list of lines proving the formula of the
        specified line from the lines specified as the assumptions of that line,
        and updating justification line numbers specified throughout the proof
        to maintain the validity of the proof. The set of allowed inference
        rules in the returned proof is the union of the rules allowed in the two
        given proofs, but the "lemma" rule that is used in the specified line in
        `main_proof` is no longer used in the corresponding lines in the
        returned proof (and thus, this "lemma" rule is used one less time in the
        returned proof than in `main_proof`).
    """
    assert main_proof.is_valid()
    assert line_number < len(main_proof.lines)
    assert str(main_proof.lines[line_number].rule) == str(lemma_proof.statement)
    assert lemma_proof.is_valid()

    s_lemma_proof = prove_specialization(lemma_proof, Proof.rule_for_line(main_proof, line_number))
    adjustment = len(lemma_proof.lines)-1

    new_lines = []
    for i in range(line_number):
        new_lines.append(main_proof.lines[i])

    for line in s_lemma_proof.lines:
        if not Proof.Line.is_assumption(line):
            rule = line.rule
            assumptions = [assumption+line_number for assumption in line.assumptions]
            new_lines.append(Proof.Line(line.formula, rule, assumptions))

        else:
            if not line.formula in main_proof.statement.assumptions:
                for earlier_line in main_proof.lines:
                    if line.formula == earlier_line.formula:
                        rule, assumptions = earlier_line.rule, earlier_line.assumptions
                        break
                new_lines.append(Proof.Line(line.formula, rule, assumptions))

            else:
                new_lines.append(line)

    for i in range(line_number+1, len(main_proof.lines)):
        if not Proof.Line.is_assumption(main_proof.lines[i]):
            formula = main_proof.lines[i].formula
            rule = main_proof.lines[i].rule
            assumptions = [assumption+adjustment if assumption >= line_number else assumption for assumption in main_proof.lines[i].assumptions]

            new_lines.append(Proof.Line(formula, rule, assumptions))

        else:
            new_lines.append(main_proof.lines[i])

    new_rules = set(lemma_proof.rules).union(set(main_proof.rules))
    return Proof(main_proof.statement, new_rules, new_lines)

    # Task 5.2a

def uses_rule(proof: Proof, rule: InferenceRule) -> bool:
    """Returns True if the given proof uses the given rule.
    """
    return any((line.rule == rule for line in proof.lines))
    # Personal task

def find_first_use_of_rule(proof: Proof, rule: InferenceRule) -> int:
    """Returns the number of the first line that uses the given inference rule.
    """
    for i in range(len(proof.lines)):
        if proof.lines[i].rule == rule:
            return i
    else:
        return False
    # Personal task

def inline_proof(main_proof: Proof, lemma_proof: Proof) -> Proof:
    """Inlines the given proof of a "lemma" inference rule into the given proof
    that uses that "lemma" rule, eliminating all usages of (any specializations
    of) that "lemma" rule in the latter proof.

    Parameters:
        main_proof: valid proof to inline into.
        lemma_proof: valid proof of one of the allowed inference rules of
            `main_proof`.

    Returns:
        A valid proof obtained from `main_proof` by inlining (an appropriate
        specialization of) `lemma_proof` in lieu of each line that specifies the
        "lemma" inference rule proved by `lemma_proof` as its justification. The
        set of allowed inference rules in the returned proof is the union of the
        rules allowed in the two given proofs but without the "lemma" rule
        proved by `lemma_proof`.
    """
    assert main_proof.is_valid(), print('\n******************** invalid proof ********************\n', main_proof, '\n***************************************************\n')
    assert lemma_proof.is_valid()

    new_rules = set(rule for rule in main_proof.rules if rule != lemma_proof.statement).union(lemma_proof.rules)

    new_proof = _inline_proof_once(main_proof, find_first_use_of_rule(main_proof, lemma_proof.statement), lemma_proof)
    while uses_rule(new_proof, lemma_proof.statement):
        new_proof = _inline_proof_once(new_proof, find_first_use_of_rule(new_proof, lemma_proof.statement), lemma_proof)
    proof1 = Proof(main_proof.statement, new_rules, new_proof.lines)
    return proof1
    # Task 5.2b

def parse_inference_rule(text:str) -> InferenceRule:
    """Converts a copied-and-pasted string from an inference rule object into an InferenceRule object"""

    assumption_str = text.split("[")[1].replace("'","").split("]")[0].split(",")
    if assumption_str == ['']:
        st_assumptions = []
    else:
        st_assumptions = [Formula.parse(i.replace("'",'').replace(" ",'')) for i in assumption_str]
    st_conclusion = Formula.parse(text.split("==> '")[1].split("'")[0])
    return InferenceRule(st_assumptions, st_conclusion)
    # Personal task

def parse_proof(text: list[str]) -> Proof:
    """Converts a copied-and-pasted string from a proof object, with triple quotes added to the start and end, into a Proof object"""

    text = text.split('\n')

    statement = parse_inference_rule(text[0])

    rules = []
    lines = []
    on_rules = True
    for i in text[1:]:
        if on_rules:
            if 'Lines' in i:
                on_rules = False
                continue
            rules.append(parse_inference_rule(i))
        elif 'QED' in i:
            break
        else:
            afternumber = ')'.join(i.split(')')[1:])
            formula = afternumber.split(' ')[1]
            try:
                formula = Formula.parse(formula)
            except AssertionError:
                print('error parsing the formula in the following line: \n', i, '\n')
            if 'Inference Rule ' in i:
                i = i.split('Inference Rule ')[1]
                if ' on lines ' in i:
                    inferencerule, assumptions = i.split(' on lines ')
                    inferencerule = parse_inference_rule(inferencerule)
                    assumptions = [int(j) for j in assumptions[:-1].split(',')]

                elif ' on line ' in i:
                    inferencerule, assumptions = i.split(' on line ')
                    inferencerule = parse_inference_rule(inferencerule)
                    assumptions = [int(j) for j in assumptions[:-1].split(',')]

                else:
                    inferencerule = parse_inference_rule(i)
                    assumptions = []
                lines.append(Proof.Line(formula, inferencerule, assumptions))

            else:
                lines.append(Proof.Line(formula))
    return Proof(statement, rules, lines)
    # Personal task

def inference_rule_to_code(rule: InferenceRule) -> str:
    """Converts an InferenceRule object into the constituent Python code, represented as a string"""

    return 'InferenceRule([' + ', '.join([f"Formula.parse('{str(i)}')" for i in rule.assumptions]) + f"], Formula.parse('{str(rule.conclusion)}'))"
    # Personal task

def proof_to_code(proof: Proof) -> str:
    """Converts a Proof object into the constituent Python code, represented as a string"""

    text = 'Proof(\n'
    statement = '\t' + inference_rule_to_code(proof.statement) + ',\n'
    rules = '\t[' + ',\n\t\t'.join([inference_rule_to_code(i) for i in proof.rules]) + '],\n'
    lines = '\t[Proof.Line(' + "),\n\t\tProof.Line(".join([f"Formula.parse('{i.formula}'), {inference_rule_to_code(i.rule)}, {list(i.assumptions)}"\
        if not i.is_assumption() else f"Formula.parse('{i.formula}')" for i in proof.lines]) + ")])"

    return text + statement + rules + lines
    # Personal task

def find_first_instance_of_formula(proof: Proof, formula: Formula) -> int:
    """Returns the line number of the first line containing the given formula"""

    for i in range(len(proof.lines)):
        if proof.lines[i].formula == formula:
            return i
    # Personal task

def convert_inference_rule_operators(inferencerule: InferenceRule, operator_function) -> InferenceRule:
    return InferenceRule([operator_function(i) for i in inferencerule.assumptions], operator_function(inferencerule.conclusion))
    # Personal task

def convert_proof_operators(proof: Proof, operator_function) -> Proof:
    statement = convert_inference_rule_operators(proof.statement, operator_function)
    rules = [convert_inference_rule_operators(i, operator_function) for i in proof.rules]

    lines = []
    for line in proof.lines:
        if line.is_assumption():
            lines.append(Proof.Line(operator_function(line.formula)))
        else:
            lines.append(Proof.Line(operator_function(line.formula), convert_inference_rule_operators(line.rule, operator_function), line.assumptions))
    return Proof(statement, rules, lines)
    # Personal task



# This file is part of the materials accompanying the book
# "Mathematical Logic through Python" by Gonczarowski and Nisan,
# Cambridge University Press. Book site: www.LogicThruPython.org
# (c) Yannai A. Gonczarowski and Noam Nisan, 2017-2022
# File name: predicates/proofs.py

"""Schemas and proofs in Predicate Logic."""

from __future__ import annotations
from typing import AbstractSet, FrozenSet, Mapping, Sequence, Tuple, Union
from mltp.logic_utils import frozen, frozendict
from mltp.propositions.semantics import is_tautology as is_propositional_tautology
from mltp.propositions.syntax import is_variable as is_propositional_variable
from .syntax import ForbiddenVariableError, Term, Formula, is_constant, is_variable, is_function, is_relation, is_equality, is_binary, is_unary, is_quantifier

#: A mapping from constant names, variable names, and relation names to
#: terms, variable names, and formulas respectively.
InstantiationMap = Mapping[str, Union[Term, str, Formula]]

@frozen
class Schema:
    """An immutable schema of predicate-logic formulas, comprised of a formula
    along with the constant names, variable names, and nullary or unary relation
    names in that formula that serve as templates. A template constant name is a
    placeholder for any term. A template variable name is a placeholder for any
    variable name. A template nullary or unary relation name is a placeholder
    for any (parametrized for a unary template relation name) predicate-logic
    formula that does not refer to any variable name in the schema (except
    possibly through its instantiated argument for a unary relation name).

    Attributes:
        formula (`~predicates.syntax.Formula`): the formula of the schema.
        templates (`~typing.FrozenSet`\\[`str`]): the constant, variable, and
            relation names from the formula that serve as templates.
    """
    formula: Formula
    templates: FrozenSet[str]

    def __init__(self, formula: Formula,
                 templates: AbstractSet[str] = frozenset()):
        """Initializes a `Schema` from its formula and template names.

        Parameters:
            formula : the formula for the schema.
            templates: the constant, variable, and relation names from the
                formula to serve as templates.
        """
        if isinstance(formula, Schema):
            self.formula = formula.formula
            self.templates = formula.templates
        else:
            for template in templates:
                assert is_constant(template) or is_variable(template) or \
                       is_relation(template)
                if is_relation(template):
                    arities = {arity for relation,arity in formula.relations() if
                               relation == template}
                    assert arities == {0} or arities == {1}
            self.formula = formula
            self.templates = frozenset(templates)

    def __repr__(self) -> str:
        """Computes a string representation of the current schema.

        Returns:
            A string representation of the current schema.
        """
        return 'Schema: ' + str(self.formula) + ' [templates: ' + \
               ('none' if len(self.templates) == 0 else
                ", ".join(sorted(self.templates))) + ']'

    def __eq__(self, other: object) -> bool:
        """Compares the current schema with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is a `Schema` object that equals the
            current schema, ``False`` otherwise.
        """
        return isinstance(other, Schema) and self.formula == other.formula and \
               self.templates == other.templates

    def __ne__(self, other: object) -> bool:
        """Compares the current schema with the given one.

        Parameters:
            other: object to compare to.

        Returns:
            ``True`` if the given object is not a `Schema` object or does not
            equal the current schema, ``False`` otherwise.
        """
        return not self == other

    def __hash__(self) -> int:
        return hash(str(self))

    class BoundVariableError(Exception):
        """Raised by `_instantiate_helper` when a variable name becomes bound
        during a schema instantiation in a way that is disallowed in that
        context.

        Attributes:
            variable_name (`str`): the variable name that became bound in a way
                that was disallowed during a schema instantiation.
            relation_name (`str`): the relation name during whose substitution
                the relevant occurrence of the variable name became bound.
        """
        variable_name: str
        relation_name: str

        def __init__(self, variable_name: str, relation_name: str):
            """Initializes a `~Schema.BoundVariableError` from the offending
            variable name and the relation name during whose substitution the
            error occurred.

            Parameters:
                variable_name: variable name that is to become bound in a way
                    that is disallowed during a schema instantiation.
                relation_name: the relation name during whose substitution the
                    relevant occurrence of the variable name is to become bound.
            """
            assert is_variable(variable_name)
            assert is_relation(relation_name)
            self.variable_name = variable_name
            self.relation_name = relation_name

    @staticmethod
    def _instantiate_helper(formula: Formula,
                            constants_and_variables_instantiation_map:
                            Mapping[str, Term],
                            relations_instantiation_map: Mapping[str, Formula],
                            bound_variables: AbstractSet[str] = frozenset()) \
            -> Formula:
        """Performs the following substitutions in the given formula:

        1. Substitute each occurrence of each constant name or variable name
           that is a key of the given constants and variables instantiation map
           with the term mapped to this name by this map.
        2. Substitute each nullary invocation of each relation name that is a    #Can I just sub in my constants and variables into my relations, then pass just my subbed relations to lower levels?
           key of the given relations instantiation map with the formula mapped
           to it by this map.
        3. For each unary invocation of each relation name that is a key of the
           given relations instantiation map, first perform all substitutions
           to the argument of this invocation (according to the given constants
           and variables instantiation map), then substitute the result for
           each occurrence of the constant name ``'_'`` in the formula mapped to
           the relation name by this map, and then substitute the result for
           this unary invocation of the relation name.

        Only name occurrences originating in the given formula are substituted
        (i.e., name occurrences originating in one of the above substitutions
        are not subjected to additional substitutions).

        Parameters:
            formula: formula in which to perform the substitutions.
            constants_and_variables_instantiation_map: mapping from constant
                names and variable names in the given formula to terms to be
                substituted for them, where the roots of terms mapped to
                variable names are variable names.
            relations_instantiation_map: mapping from nullary and unary relation
                names in the given formula to formulas to be substituted for
                them, where formulas to be substituted for unary relation names
                are parametrized by the constant name ``'_'``.
            bound_variables: variable names to be treated as bound (see below).

        Returns:
            The result of all substitutions.

        Raises:
            BoundVariableError: if one of the following occurs when substituting
                an invocation of a relation name:

                1. A free occurrence of a variable name in the formula
                   mapped to the relation name by the given relations
                   instantiation map is in `bound_variables`.
                2. A free occurrence of a variable name in the formula
                   mapped to the relation name by the given relations
                   instantiation map becomes bound by a quantification in
                   the given formula after all variable
                   names in the given formula have been substituted.
                3. For a unary invocation: a variable name that is in the
                   argument to that invocation after all substitutions have been
                   applied to this argument becomes bound by a quantification
                   in the formula mapped to the relation name by the given
                   relations instantiation map.

        Examples:
            The following succeeds:

            >>> Schema._instantiate_helper(
            ...     Formula.parse('Ax[(Q(c)->R(x))]'), {'x': Term('w')},
            ...     {'Q': Formula.parse('y=_')}, {'x', 'z'})
            Aw[(y=c->R(w))]

            however the following fails since ``'Q(c)'`` is to be substituted
            with ``'y=c'`` while ``'y'`` is specified to be treated as bound:

            >>> Schema._instantiate_helper(
            ...     Formula.parse('Ax[(Q(c)->R(x))]'), {},
            ...     {'Q': Formula.parse('y=_')}, {'x', 'y', 'z'})
            Traceback (most recent call last):
              ...
            predicates.proofs.Schema.BoundVariableError: ('y', 'Q')

            and the following fails since as ``'Q(c)'`` is to be substituted
            with ``'y=c'``, ``'y'`` is to become bound by the quantification
            ``'Ay'``:

            >>> Schema._instantiate_helper(
            ...     Formula.parse('Ax[(Q(c)->R(x))]'), {'x': Term('y')},
            ...     {'Q': Formula.parse('y=_')})
            Traceback (most recent call last):
              ...
            predicates.proofs.Schema.BoundVariableError: ('y', 'Q')

            The following succeeds:

            >>> Schema._instantiate_helper(
            ...     Formula.parse('Ax[(Q(c)->R(x))]'),
            ...     {'c': Term.parse('plus(d,x)')},
            ...     {'Q': Formula.parse('Ey[y=_]')})
            Ax[(Ey[y=plus(d,x)]->R(x))]

            however the following fails since as ``'_'`` is to be substituted
            with ``'plus(d,y)'`` in ``'Ey[y=_]'``, the ``'y'`` in
            ``'plus(d,y)'`` is to become bound by the quantification ``'Ey'``:

            >>> Schema._instantiate_helper(
            ...     Formula.parse('Ax[(Q(c)->R(x))]'),
            ...     {'c': Term.parse('plus(d,y)')},
            ...     {'Q': Formula.parse('Ey[y=_]')})
            Traceback (most recent call last):
              ...
            predicates.proofs.Schema.BoundVariableError: ('y', 'Q')
        """
        for construct in constants_and_variables_instantiation_map:
            assert is_constant(construct) or is_variable(construct)
            if is_variable(construct):
                assert is_variable(
                    constants_and_variables_instantiation_map[construct].root)
        for relation in relations_instantiation_map:
            assert is_relation(relation)
        for variable in bound_variables:
            assert is_variable(variable)
        root = formula.root

        if is_unary(root):
            try:
                first = Schema._instantiate_helper(formula.first,
                                constants_and_variables_instantiation_map,
                                relations_instantiation_map,
                                bound_variables)
            except ForbiddenVariableError as e:
                raise Schema.BoundVariableError(e.variable_name, formula.first.root)
            formula = Formula(root, first)

        elif is_binary(root):
            try:
                first = Schema._instantiate_helper(formula.first,
                                constants_and_variables_instantiation_map,
                                relations_instantiation_map,
                                bound_variables)
                second = Schema._instantiate_helper(formula.second,
                                constants_and_variables_instantiation_map,
                                relations_instantiation_map,
                                bound_variables)
            except ForbiddenVariableError as e:
                raise Schema.BoundVariableError(e.variable_name, formula.first.root)
            formula = Formula(root, first, second)

        elif is_relation(root):
            relation_name, arity = formula.relations().pop()
            if arity == 1: #parametrized
                formula = formula.substitute(constants_and_variables_instantiation_map)
                parameter = formula.arguments[0]
                if relation_name in relations_instantiation_map:
                    replacement_template = relations_instantiation_map[relation_name]

                    if sinners := replacement_template.free_variables() & bound_variables:
                        raise Schema.BoundVariableError(sinners.pop(), relation_name)

                    try:
                        formula = replacement_template.substitute({'_':parameter})
                    except ForbiddenVariableError as e:
                        raise Schema.BoundVariableError(e.variable_name, relation_name)

            else: #parameterless
                if relation_name in relations_instantiation_map:
                    replacement_template = relations_instantiation_map[relation_name]
                    if sinners := replacement_template.free_variables() & bound_variables:
                        raise Schema.BoundVariableError(sinners.pop(), relation_name)
                    formula = replacement_template

        elif is_equality(root):
            formula = formula.substitute(constants_and_variables_instantiation_map)

        elif is_quantifier(root):
            variable = formula.variable
            if variable in constants_and_variables_instantiation_map:
                bound_variables = bound_variables - {variable}
                variable = str(constants_and_variables_instantiation_map[variable])
            bound_variables = bound_variables | {variable}
            formula = Schema._instantiate_helper(formula.statement,
                            constants_and_variables_instantiation_map,
                            relations_instantiation_map,
                            bound_variables)
            formula = Formula(root, variable, formula)
        return formula
        # Task 9.3

    def instantiate(self, instantiation_map: InstantiationMap) -> \
            Union[Formula, None]:
        """Instantiates the current schema according to the given map from
        templates of the current schema to expressions.

        Parameters:
            instantiation_map: mapping from templates of the current schema to
                expressions of the type for which they serve as placeholders.
                That is, constant names are mapped to terms, variable names are
                mapped to variable names (strings), and relation names are
                mapped to formulas where unary relation names are mapped to
                formulas parametrized by the constant name ``'_'``.

        Returns:
            The predicate-logic formula obtained by applying the substitutions
            specified by the given map to the formula of the current schema:

            1. Each occurrence in the formula of the current schema of each
               template constant name specified in the given map is substituted
               with the term to which that template constant name is mapped.
            2. Each occurrence in the formula of the current schema of each
               template variable name specified in the given map is substituted
               with the variable name to which that template variable name is
               mapped.
            3. Each nullary invocation in the formula of the current schema of
               each template relation name specified in the given map is
               substituted with the formula to which that template relation name
               is mapped.
            4. Each unary invocation in the formula of the current schema of
               each template relation name specified in the given map is
               substituted with the formula to which that template relation name
               is mapped, in which each occurrence of the constant name ``'_'``
               is substituted with  the instantiated argument of that invocation
               of the template relation name (that is, the term that results
               from instantiating the argument of that invocation by performing
               all the specified substitutions on it).

            ``None`` is returned if one of the keys of the given map is not a
            template of the current schema or if one of the following occurs
            when substituting an invocation of a template relation name:

            1. A free occurrence of a variable name in the formula substituted
               for the template relation name becomes bound by a quantification
               in the instantiated schema formula, except if the template
               relation name is unary and this free occurrence originates in the
               instantiated argument of the invocation of the template relation
               name.
            2. For a unary invocation: a variable name in the instantiated
               argument of that invocation becomes bound by a quantification in
               the formula that is substituted for the invocation of the
               template relation name.

        Examples:
            >>> s = Schema(Formula.parse('(Q(c1,c2)->(R(c1)->R(c2)))'),
            ...            {'c1', 'c2', 'R'})
            >>> s.instantiate({'c1': Term.parse('plus(x,1)'),
            ...                'R': Formula.parse('Q(_,y)')})
            (Q(plus(x,1),c2)->(Q(plus(x,1),y)->Q(c2,y)))
            >>> s.instantiate({'c1': Term.parse('plus(x,1)'),
            ...                'c2': Term.parse('c1'),
            ...                'R': Formula.parse('Q(_,y)')})
            (Q(plus(x,1),c1)->(Q(plus(x,1),y)->Q(c1,y)))

            >>> s = Schema(Formula.parse('(P()->P())'), {'P'})
            >>> s.instantiate({'P': Formula.parse('plus(a,b)=c')})
            (plus(a,b)=c->plus(a,b)=c)

            For the following schema:

            >>> s = Schema(Formula.parse('(Q(d)->Ax[(R(x)->Q(f(c)))])'),
            ...            {'R', 'Q', 'x', 'c'})

            the following succeeds:

            >>> s.instantiate({'R': Formula.parse('_=0'),
            ...                'Q': Formula.parse('x=_'),
            ...                'x': 'w'})
            (x=d->Aw[(w=0->x=f(c))])

            however, the following returns ``None`` because ``'d'`` is not a
            template of the schema:

            >>> s.instantiate({'R': Formula.parse('_=0'),
            ...                'Q': Formula.parse('x=_'),
            ...                'x': 'w',
            ...                'd': Term('z')})

            and the following returns ``None`` because ``'z'`` that is free in
            the assignment to ``'Q'`` is to become bound by a quantification in
            the instantiated schema formula:

            >>> s.instantiate({'R': Formula.parse('_=0'),
            ...                'Q': Formula.parse('s(z)=_'),
            ...                'x': 'z'})

            and the following returns ``None`` because ``'y'`` in the
            instantiated argument ``'f(plus(a,y))'`` of the second invocation of
            ``'Q'`` is to become bound by the quantification in the formula
            substituted for ``'Q'``:

            >>> s.instantiate({'R': Formula.parse('_=0'),
            ...                'Q': Formula.parse('Ay[s(y)=_]'),
            ...                'c': Term.parse('plus(a,y)')})
        """
        for construct in instantiation_map:
            if is_variable(construct):
                assert is_variable(instantiation_map[construct])
            elif is_constant(construct):
                assert isinstance(instantiation_map[construct], Term)
            else:
                assert is_relation(construct)
                assert isinstance(instantiation_map[construct], Formula)

        constants_and_variables_instantiation_map = {}
        relations_instantiation_map = {}
        for key in instantiation_map:
            if not key in self.templates:
                return None
            elif is_variable(key):
                assert isinstance(instantiation_map[key], str)
                constants_and_variables_instantiation_map[key] = Term(instantiation_map[key])
            elif is_constant(key):
                assert isinstance(instantiation_map[key], Term)
                constants_and_variables_instantiation_map[key] = instantiation_map[key]
            elif is_relation(key):
                assert isinstance(instantiation_map[key], Formula)
                relations_instantiation_map[key] = instantiation_map[key]

        if not (constants_and_variables_instantiation_map or relations_instantiation_map):
            return self.formula

        bound_variables = set()
        try:
            return Schema._instantiate_helper(self.formula, constants_and_variables_instantiation_map, relations_instantiation_map, bound_variables)
        except Schema.BoundVariableError as e:
            print(f'BoundVariableError: {e.variable_name} is bound within the relation {e.relation_name}')
            return None
        # Task 9.4

    @staticmethod
    def parse(string: str) -> Schema:
        """Parses the given valid string representation into a Schema.

        Parameters:
            string: string to parse.

        Returns:
            A schema whose standard string representation is the given string.
        """
        formula, templates = string.split('Schema: ')[1].split(' ', maxsplit = 1)
        formula = Formula.parse(formula)
        templates = set(templates.split(': ')[1][:-1].split(', '))
        if templates == {'none'}:
            templates = {}
        return Schema(formula, templates)
        # personal task

@frozen
class Proof:
    """An immutable deductive proof in Predicate Logic, comprised of a list of
    assumptions/axioms, a conclusion, and a list of lines that prove the
    conclusion from (instances of) these assumptions/axioms and from
    tautologies, via the Modus Ponens (MP) and Universal Generalization (UG)
    inference rules.

    Attributes:
        assumptions (`~typing.FrozenSet`\\[`Schema`]): the assumption/axioms of
            the proof.
        conclusion (`~predicates.syntax.Formula`): the conclusion of the proof.
        lines (`~typing.Tuple`\\[`Line`\]): the lines of the proof.
    """
    assumptions: FrozenSet[Schema]
    conclusion: Formula
    lines: Tuple[Proof.Line, ...]

    def __init__(self, assumptions: AbstractSet[Schema], conclusion: Formula,
                 lines: Sequence[Proof.Line]):
        """Initializes a `Proof` from its assumptions/axioms, conclusion,
        and lines.

        Parameters:
            assumptions: the assumption/axioms for the proof.
            conclusion: the conclusion for the proof.
            lines: the lines for the proof.
        """
        self.assumptions = frozenset(assumptions)
        self.conclusion = conclusion
        self.lines = tuple(lines)

    @frozen
    class AssumptionLine:
        """An immutable proof line justified as an instance of an
        assumption/axiom.

        Attributes:
            formula (`~predicates.syntax.Formula`): the formula justified by the
                line.
            assumption (`Schema`): the assumption/axiom that instantiates the
                formula.
            instantiation_map (`~typing.Mapping`\\[`str`, `~typing.Union`\\[`~predicates.syntax.Term`, `str`, `~predicates.syntax.Formula`]]):
                the mapping instantiating the formula from the assumption/axiom.
        """
        formula: Formula
        assumption: Schema
        instantiation_map: InstantiationMap

        def __init__(self, formula: Formula, assumption: Schema,
                     instantiation_map: InstantiationMap):
            """Initializes an `~Proof.AssumptionLine` from its formula, its
            justifying assumption/axiom, and its instantiation map from the
            justifying assumption/axiom.

            Parameters:
                formula: the formula to be justified by the line.
                assumption: the assumption/axiom that instantiates the formula.
                instantiation_map: the mapping instantiating the formula from
                    the assumption/axiom.
            """
            for construct in instantiation_map:
                if is_variable(construct):
                    assert is_variable(instantiation_map[construct])
                elif is_constant(construct):
                    assert isinstance(instantiation_map[construct], Term)
                else:
                    assert is_relation(construct)
                    assert isinstance(instantiation_map[construct], Formula)
            self.formula = formula
            self.assumption = assumption
            self.instantiation_map = frozendict(instantiation_map)

        def __repr__(self) -> str:
            """Computes a string representation of the current line.

            Returns:
                A string representation of the current line.
            """
            return str(self.formula) + "    (Assumption " + \
                   str(self.assumption) + " instantiated with " + \
                   str(self.instantiation_map) + ")"

        def is_valid(self, assumptions: AbstractSet[Schema],
                     lines: Sequence[Proof.Line], line_number: int) -> bool:
            """Checks if the current line is validly justified in the context of
            the specified proof.

            Parameters:
                assumptions: assumptions/axioms of the proof.
                lines: lines of the proof.
                line_number: line number of the current line in the given lines.

            Returns:
                ``True`` if the assumption/axiom of the current line is an
                assumption/axiom of the specified proof and if the formula
                justified by the current line is a valid instance of this
                assumption/axiom via the instantiation map of the current line,
                ``False`` otherwise.
            """
            assert line_number < len(lines) and lines[line_number] is self
            assumption_is_valid = self.assumption in assumptions
            formula_is_valid = self.formula == self.assumption.instantiate(self.instantiation_map)
            if assumption_is_valid and formula_is_valid:
                return True
            else:
                print(f'Invalid line [{line_number, self}]\nassumption is valid: {assumption_is_valid}\nformula is valid: {formula_is_valid}')
                return False
            # Task 9.5

        @staticmethod
        def parse(string: str) -> Proof.AssumptionLine:
            """Parses the given valid string representation into an AssumptionLine.

            Parameters:
                string: string to parse.

            Returns:
                An AssumptionLine whose standard string representation is the given string.
            """
            formula = Formula.parse(string.split(') ', maxsplit=1)[1].split('    (')[0])
            assumption, instantiation_map = string.split('(Assumption ')[1].split(' instantiated with ')
            assumption = Schema.parse(assumption)
            instantiation_map = instantiation_map[1:-2].replace("'",'').replace(' ','').split(',')
            if instantiation_map and instantiation_map[0]:
                instantiation_map = dict(i.split(':') for i in instantiation_map)
                for key in instantiation_map:
                    if is_constant(key):
                        instantiation_map[key] = Term.parse(instantiation_map[key])
                    elif is_relation(key):
                        instantiation_map[key] = Formula.parse(instantiation_map[key])
            else:
                instantiation_map = {}
            return Proof.AssumptionLine(formula, assumption, instantiation_map)
            # personal task

    @frozen
    class MPLine:
        """An immutable proof line justified by the Modus Ponens (MP) inference
        rule.

        Attributes:
            formula (`~predicates.syntax.Formula`): the formula justified by the
                line.
            antecedent_line_number (`int`): the line number of the antecedent of
                the MP inference justifying the line.
            conditional_line_number (`int`): the line number of the conditional
                of the MP inference justifying the line.
        """
        formula: Formula
        antecedent_line_number: int
        conditional_line_number: int

        def __init__(self, formula: Formula, antecedent_line_number: int,
                     conditional_line_number: int):
            """Initializes a `~Proof.MPLine` from its formula and line numbers
            of the antecedent and conditional of the MP inference justifying it.

            Parameters:
                formula: the formula to be justified by the line.
                antecedent_line_number: the line number of the antecedent of the
                    MP inference to justify the line.
                conditional_line_number: the line number of the conditional of
                    the MP inference to justify the line.
            """
            self.formula = formula
            self.antecedent_line_number = antecedent_line_number
            self.conditional_line_number = conditional_line_number

        def __repr__(self) -> str:
            """Computes a string representation of the current line.

            Returns:
                A string representation of the current line.
            """
            return str(self.formula) + "    (MP from lines " + \
                   str(self.antecedent_line_number) + " and " + \
                   str(self.conditional_line_number) + ")"

        def is_valid(self, assumptions: AbstractSet[Schema],
                     lines: Sequence[Proof.Line], line_number: int) -> bool:
            """Checks if the current line is validly justified in the context of
            the specified proof.

            Parameters:
                assumptions: assumptions/axioms of the proof.
                lines: lines of the proof.
                line_number: line number of the current line in the given lines.

            Returns:
                ``True`` if the formula of the line from the given lines whose
                number is the conditional line number justifying the current
                line is ``'(``\ `antecedent`\ ``->``\ `consequent`\ ``)'``,
                where `antecedent` is the formula of the line from the given
                lines whose number is the antecedent line number justifying the
                current line and `consequent` is the formula justified by the
                current line; ``False`` otherwise.
            """
            assert line_number < len(lines) and lines[line_number] is self
            formula_is_valid = lines[self.conditional_line_number].formula == Formula('->', lines[self.antecedent_line_number].formula, self.formula)
            if formula_is_valid and line_number > self.antecedent_line_number and line_number > self.conditional_line_number:
                return True
            else:
                if not formula_is_valid:
                    print(f'Invalid formula {self.formula}.\nAntecedent is   {lines[self.antecedent_line_number].formula},\nconditional is  {lines[self.conditional_line_number].formula}\n')
                if not line_number > self.antecedent_line_number:
                    print(f'\nline is after antecedent: {line_number > self.antecedent_line_number}')
                if not line_number > self.conditional_line_number:
                    print(f'\nline is after consequent: {line_number > self.conditional_line_number}')

                return False
            # Task 9.6

        @staticmethod
        def parse(string: str) -> Proof.MPLine:
            """Parses the given valid string representation into an MPLine.

            Parameters:
                string: string to parse.

            Returns:
                An MPLine whose standard string representation is the given string.
            """
            formula = Formula.parse(string.split(') ', maxsplit=1)[1].split('    (', maxsplit=1)[0])
            antecedent, conditional = string.split(' lines ')[1][:-1].split(' and ')
            return Proof.MPLine(formula, int(antecedent), int(conditional))
            # personal task

    @frozen
    class UGLine:
        """An immutable proof line justified by the Universal Generalization
        (UG) inference rule.

        Attributes:
            formula (`~predicates.syntax.Formula`): the formula justified by the
                line.
            nonquantified_line_number (`int`): the line number of the statement
                quantified by the formula.
        """
        formula: Formula
        nonquantified_line_number: int

        def __init__(self, formula: Formula, nonquantified_line_number: int):
            """Initializes a `~Proof.UGLine` from its formula and line number of
            the statement quantified by the formula.

            Parameters:
                formula: the formula to be justified by the line.
                nonquantified_line_number: the line number of the statement
                    quantified by the formula.
            """
            self.formula = formula
            self.nonquantified_line_number = nonquantified_line_number

        def __repr__(self) -> str:
            """Computes a string representation of the current line.

            Returns:
                A string representation of the current line.
            """
            return str(self.formula) + "    (UG of line " + \
                   str(self.nonquantified_line_number) + ")"

        def is_valid(self, assumptions: AbstractSet[Schema],
                     lines: Sequence[Proof.Line], line_number: int) -> bool:
            """Checks if the current line is validly justified in the context of
            the specified proof.

            Parameters:
                assumptions: assumptions/axioms of the proof.
                lines: lines of the proof.
                line_number: line number of the current line in the given lines.

            Returns:
                ``True`` if the formula of the current line is of the form
                `'Ax[nonquantified]'`, where
                `nonquantified` is the formula of the line from the given lines
                whose number is the nonquantified line number justifying the
                current line and `x` is any variable name; ``False`` otherwise.
            """
            assert line_number < len(lines) and lines[line_number] is self
            assumption_is_prior = line_number > self.nonquantified_line_number
            if quantifier_is_valid := self.formula.root == 'A':
                variable_is_valid = is_variable(self.formula.variable)
                formula_is_valid = self.formula.statement == lines[self.nonquantified_line_number].formula
                if formula_is_valid and variable_is_valid and assumption_is_prior:
                    return True
                else:
                    print(f'Invalid line [{self}]\nassumption is prior: {assumption_is_prior}\nquantifier_is_valid: {quantifier_is_valid}\nvariable_is_valid: {variable_is_valid}\nformula_is_valid: {formula_is_valid}')
                    return False
            print(f'Invalid line [{self}]\nassumption is prior: {assumption_is_prior}\nquantifier_is_valid: {quantifier_is_valid}')
            return False
            # Task 9.7

        @staticmethod
        def parse(string: str) -> Proof.UGLine:
            """Parses the given valid string representation into a UGLine.

            Parameters:
                string: string to parse.

            Returns:
                A UGLine whose standard string representation is the given string.
            """
            prefix, suffix = string[:-1].split('    (UG of line ')
            formula = Formula.parse(prefix.split(') ',maxsplit=1)[1])
            return Proof.UGLine(formula, int(suffix))
            # personal task

    @frozen
    class TautologyLine:
        """An immutable proof line justified as a tautology.

        Attributes:
            formula (`~predicates.syntax.Formula`): the formula justified by the
                line.
        """
        formula: Formula

        def __init__(self, formula: Formula):
            """Initializes a `~Proof.TautologyLine` from its formula.

            Parameters:
                formula: the formula to be justified by the line.
            """
            self.formula = formula

        def __repr__(self) -> str:
            """Computes a string representation of the current line.

            Returns:
                A string representation of the current line.
            """
            return str(self.formula) + "    (Tautology)"

        def is_valid(self, assumptions: AbstractSet[Schema],
                     lines: Sequence[Proof.Line], line_number: int) -> bool:
            """Checks if the current line is validly justified in the context of
            the specified proof.

            Parameters:
                assumptions: assumptions/axioms of the proof.
                lines: lines of the proof.
                line_number: line number of the current line in the given lines.

            Returns:
                ``True`` if the formula justified by the current line is a
                (predicate-logic) tautology, ``False`` otherwise.
            """
            assert line_number < len(lines) and lines[line_number] is self
            prop_skeleton = self.formula.propositional_skeleton()[0]
            if is_tautology := is_propositional_tautology(prop_skeleton):
                return True
            else:
                print(f'Invalid line: {self}\nThe formula is not a predicate logic tautology because {prop_skeleton} is not a propositional logic tautology')
                return False
            # Task 9.9

        @staticmethod
        def parse(string: str) -> Proof.TautologyLine:
            """Parses the given valid string representation into a TautologyLine.

            Parameters:
                string: string to parse.

            Returns:
                A TautologyLine whose standard string representation is the given string.
            """
            return Proof.TautologyLine(Formula.parse(string.split(') ',maxsplit=1)[1].split('    (')[0]))
            # personal task

    #: An immutable proof line.
    Line = Union[AssumptionLine, MPLine, UGLine, TautologyLine]

    def __repr__(self) -> str:
        """Computes a string representation of the current proof.

        Returns:
            A string representation of the current proof.
        """
        r = 'Proof of ' + str(self.conclusion) + ' from assumptions/axioms:\n'
        for assumption in self.assumptions:
            r += '  '  + str(assumption) + '\n'
        r += 'Lines:\n'
        for i in range(len(self.lines)):
            r += ('%3d) ' % i) + str(self.lines[i]) + '\n'
        r += 'QED\n'
        return r

    def is_valid(self) -> bool:
        """Checks if the current proof is a valid proof of its claimed
        conclusion from (instances of) its assumptions/axioms.

        Returns:
            ``True`` if the current proof is a valid proof of its claimed
            conclusion from (instances of) its assumptions/axioms, ``False``
            otherwise.
        """
        if len(self.lines) == 0 or self.lines[-1].formula != self.conclusion:
            return False
        for line_number in range(len(self.lines)):
            if not self.lines[line_number].is_valid(self.assumptions,
                                                    self.lines, line_number):
                return False
        return True

    @staticmethod
    def parse(string: str) -> Proof:
        """Parses the given valid string representation into a Proof.

        Parameters:
            string: string to parse.

        Returns:
            A proof whose standard string representation is the given string.
        """
        conclusion, suffix = string.split('\n', maxsplit=1)
        conclusion = Formula.parse(conclusion.split('Proof of ')[1].split(' from')[0])

        assumptions, suffix = suffix.split('Lines:\n')
        if assumptions:
            assumptions = [Schema.parse(i) for i in assumptions.replace('  ','').split('\n') if i]

        lines = [i for i in suffix.split('\n')[:-1]]
        new_lines = []
        for line in lines:
            if 'Assumption' in line:
                new_lines.append(Proof.AssumptionLine.parse(line))
            elif 'MP from lines' in line:
                new_lines.append(Proof.MPLine.parse(line))
            elif 'Tautology' in line:
                new_lines.append(Proof.TautologyLine.parse(line))
            elif 'UG of line' in line:
                new_lines.append(Proof.UGLine.parse(line))
        return Proof(assumptions, conclusion, new_lines)
        # personal task

    def cited_by(self, line_number) -> list:
        """Returns a list of lines cited by the given line"""

        line = self.lines[line_number]
        if isinstance(line, Proof.UGLine):
            return [line.nonquantified_line_number]
        elif isinstance(line, Proof.MPLine):
            return [line.antecedent_line_number, line.conditional_line_number]
        else:
            return []

    def uncite_duplicate_lines(self) -> list:
        """Adjusts the assumptions of any line that cites a previous line with a duplicate
       so that its assumptions cite the first occurence of that formula."""

        formulae = [line.formula for line in self.lines]
        duplicates = []
        first_occurrences = {}
        for line_number in range(len(formulae)):
            formula = formulae[line_number]
            if formula in formulae[:line_number]:
                duplicates.append(line_number)
                if formula not in first_occurrences:
                    first_occurrences[formula] = line_number

        replacement_dict = {}
        for line_number in duplicates:
            replacement_dict[line_number] = first_occurrences[self.lines[line_number].formula]

        adjusted = []
        for line in self.lines:
            if isinstance(line, Proof.UGLine):
                if line.nonquantified_line_number in replacement_dict:
                    adjusted.append(Proof.UGLine(line.formula,
                                              replacement_dict[line.nonquantified_line_number]))
                else:
                    adjusted.append(line)
            elif isinstance(line, Proof.MPLine):
                if line.antecedent_line_number in replacement_dict:
                    antecedent = replacement_dict[line.antecedent_line_number]
                else:
                    antecedent = line.antecedent_line_number

                if line.conditional_line_number in replacement_dict:
                    conditional = replacement_dict[line.conditional_line_number]
                else:
                    conditional = line.conditional_line_number

                adjusted.append(Proof.MPLine(line.formula,
                                              antecedent,
                                              conditional))
            else:
                adjusted.append(line)
        return Proof(self.assumptions, self.conclusion, adjusted)
        # personal task

    def uncited_lines(self) -> set:
        """Returns a list of all lines that are not cited by the conclusion, or
        not cited by any line that is cited by the conclusion."""
        #add conclusion
        cited = {len(self.lines)-1}
        temp = list(cited)
        #starting with conclusion, add cited lines to temp and cited. Follow up by finding lines cited by temp, and reset temp.
        while temp:
            current = temp.pop()
            new = self.cited_by(current)
            for i in new:
                temp.insert(0,i)
                cited.add(i)

        return set(range(len(self.lines)))-cited
        # personal task

    def remove_uncited_lines(self) -> Proof:
        """Returns a new Proof of the same conclusion, with unncessary lines removed."""
        lines_to_remove = self.uncited_lines()
        lines_to_keep = sorted(set(i for i in range(len(self.lines))) - set(lines_to_remove))

        line_dict = {lines_to_keep[i]:i for i in range(len(lines_to_keep))}

        new_lines = []
        for line in range(len(self.lines)):
            if line not in lines_to_remove:
                new_lines.append(self.lines[line])
                line = self.lines[line]

        renumbered_lines = []
        for line in new_lines:
            if isinstance(line, Proof.UGLine):
                renumbered_lines.append(Proof.UGLine(line.formula,
                                              line_dict[line.nonquantified_line_number]))

            elif isinstance(line, Proof.MPLine):
                renumbered_lines.append(Proof.MPLine(line.formula,
                                              line_dict[line.antecedent_line_number],
                                              line_dict[line.conditional_line_number]))
            else:
                renumbered_lines.append(line)
        return Proof(self.assumptions, self.conclusion, renumbered_lines)
        # personal task

    def clean(self, report=False) -> Proof:
        """Returns a proof with no duplicated or unnecessary lines."""
        proof = self.uncite_duplicate_lines()
        proof = self.remove_uncited_lines()
        assert self.is_valid() == proof.is_valid()
        if report:
            print('Removed ', len(self.lines)-len(proof.lines), ' lines')
        return proof
        # personal task

from mltp.propositions.proofs import Proof as PropositionalProof, \
                                InferenceRule as PropositionalInferenceRule, \
                                SpecializationMap as \
                                PropositionalSpecializationMap
from mltp.propositions.axiomatic_systems import AXIOMATIC_SYSTEM as \
                                           PROPOSITIONAL_AXIOMATIC_SYSTEM, \
                                           MP, I0, I1, D, I2, N, NI, NN, R
from mltp.propositions.tautology import prove_tautology as \
                                   prove_propositional_tautology

# Schema equivalents of the propositional-logic axioms for implication and
# negation

#: Schema equivalent of the propositional-logic self implication axiom
#: `~propositions.axiomatic_systems.I0`.
I0_SCHEMA = Schema(Formula.parse('(P()->P())'), {'P'})
#: Schema equivalent of the propositional-logic implication introduction (right)
#: axiom `~propositions.axiomatic_systems.I1`.
I1_SCHEMA = Schema(Formula.parse('(Q()->(P()->Q()))'), {'P', 'Q'})
#: Schema equivalent of the propositional-logic self-distribution of implication
#: axiom `~propositions.axiomatic_systems.D`.
D_SCHEMA = Schema(Formula.parse(
    '((P()->(Q()->R()))->((P()->Q())->(P()->R())))'), {'P', 'Q', 'R'})
#: Schema equivalent of the propositional-logic implication introduction (left)
#: axiom `~propositions.axiomatic_systems.I2`.
I2_SCHEMA = Schema(Formula.parse('(~P()->(P()->Q()))'), {'P', 'Q'})
#: Schema equivalent of the propositional-logic converse contraposition axiom
#: `~propositions.axiomatic_systems.N`.
N_SCHEMA  = Schema(Formula.parse('((~Q()->~P())->(P()->Q()))'), {'P', 'Q'})
#: Schema equivalent of the propositional-logic negative-implication
#: introduction axiom `~propositions.axiomatic_systems.NI`.
NI_SCHEMA = Schema(Formula.parse('(P()->(~Q()->~(P()->Q())))'), {'P', 'Q'})
#: Schema equivalent of the propositional-logic double-negation introduction
#: axiom `~propositions.axiomatic_systems.NN`.
NN_SCHEMA = Schema(Formula.parse('(P()->~~P())'), {'P'})
#: Schema equivalent of the propositional-logic resolution axiom
#: `~propositions.axiomatic_systems.R`.
R_SCHEMA  = Schema(Formula.parse(
    '((Q()->P())->((~Q()->P())->P()))'), {'P', 'Q'})

#: Schema system equivalent of the axioms of the propositional-logic large
#: axiomatic system for implication and negation
#: `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`.
PROPOSITIONAL_AXIOMATIC_SYSTEM_SCHEMAS = {I0_SCHEMA, I1_SCHEMA, D_SCHEMA,
                                          I2_SCHEMA, N_SCHEMA, NI_SCHEMA,
                                          NN_SCHEMA, R_SCHEMA}

#: Mapping from propositional-logic axioms for implication and negation to their
#: schema equivalents.
PROPOSITIONAL_AXIOM_TO_SCHEMA = {
        I0: I0_SCHEMA, I1: I1_SCHEMA, D: D_SCHEMA, I2: I2_SCHEMA, N: N_SCHEMA,
        NI: NI_SCHEMA, NN: NN_SCHEMA, R: R_SCHEMA}

def _axiom_specialization_map_to_schema_instantiation_map(
        propositional_specialization_map: PropositionalSpecializationMap,
        substitution_map: Mapping[str, Formula]) -> Mapping[str, Formula]:
    """Composes the given propositional-logic specialization map, specifying the
    transformation from a propositional-logic axiom to a specialization of it,
    and the given substitution map, specifying the transformation from that
    specialization (as a propositional skeleton) to a predicate-logic formula,
    into an instantiation map specifying how to instantiate the schema
    equivalent of that axiom into the same predicate-logic formula.

    Parameters:
        propositional_specialization_map: mapping specifying how some
            propositional-logic axiom `axiom` (which is not specified) from
            `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM` specializes into
            some specialization `specialization` (which is also not specified),
            and containing no constants or operators beyond ``'~'``, ``'->'``,
            ``'|'``, and ``'&'``.
        substitution_map: mapping from each propositional variable name of
            `specialization` to a predicate-logic formula.

    Returns:
        An instantiation map for instantiating the schema equivalent of `axiom`
        into the predicate-logic formula obtained from its propositional
        skeleton `specialization` by the given substitution map.

    Examples:
        >>> _axiom_specialization_map_to_schema_instantiation_map(
        ...     {'p': PropositionalFormula.parse('(z1->z2)'),
        ...      'q': PropositionalFormula.parse('~z1')},
        ...     {'z1': Formula.parse('Ax[(x=5&M())]'),
        ...      'z2': Formula.parse('R(f(8,9))')})
        {'P': (Ax[(x=5&M())]->R(f(8,9))), 'Q': ~Ax[(x=5&M())]}
    """
    for variable in propositional_specialization_map:
        assert is_propositional_variable(variable)
        for operator in propositional_specialization_map[variable].operators():
            assert is_unary(operator) or is_binary(operator)
    for variable in substitution_map:
        assert is_propositional_variable(variable)

    new_dict = {}
    for key in propositional_specialization_map:
        specialization = propositional_specialization_map[key]
        pred_formula = Formula.from_propositional_skeleton(specialization, substitution_map)
        new_dict[key.upper()] = pred_formula
    return new_dict
    # Task 9.11a

def _prove_from_skeleton_proof(formula: Formula,
                               skeleton_proof: PropositionalProof,
                               substitution_map: Mapping[str, Formula]) -> \
        Proof:
    """Translates the given proof of a propositional skeleton of the given
    predicate-logic formula into a proof of that predicate-logic formula.

    Parameters:
        formula: predicate-logic formula to prove.
        skeleton_proof: valid propositional-logic proof of a propositional
            skeleton of the given formula, from no assumptions and via
            `~propositions.axiomatic_systems.AXIOMATIC_SYSTEM`, and containing
            no constants or operators beyond ``'~'``, ``'->'``, ``'|'``, and
            ``'&'``.
        substitution_map: mapping from each propositional variable name of the
            propositional skeleton of the given formula that is proven in the
            given proof to the respective predicate-logic subformula of the
            given formula.

    Returns:
        A valid predicate-logic proof of the given formula from the axioms
        `PROPOSITIONAL_AXIOMATIC_SYSTEM_SCHEMAS` via only assumption lines and
        MP lines.
    """
    skeleton_proof = skeleton_proof.clean()
    assert len(skeleton_proof.statement.assumptions) == 0 and \
           skeleton_proof.rules.issubset(PROPOSITIONAL_AXIOMATIC_SYSTEM) and \
           skeleton_proof.is_valid()
    assert Formula.from_propositional_skeleton(
        skeleton_proof.statement.conclusion, substitution_map) == formula
    for line in skeleton_proof.lines:
        for operator in line.formula.operators():
            assert is_unary(operator) or is_binary(operator)

    lines = []
    for line in skeleton_proof.lines:
        new_formula = Formula.from_propositional_skeleton(line.formula, substitution_map)

        if line.rule in PROPOSITIONAL_AXIOM_TO_SCHEMA:
            new_rule = PROPOSITIONAL_AXIOM_TO_SCHEMA[line.rule]

            propositional_specialization_map = PropositionalInferenceRule._formula_specialization_map(line.rule.conclusion,line.formula)
            instantiation_map = _axiom_specialization_map_to_schema_instantiation_map(propositional_specialization_map, substitution_map)

            lines.append(Proof.AssumptionLine(new_formula, new_rule, instantiation_map))

        else:
            lines.append(Proof.MPLine(new_formula, *line.assumptions))

    return Proof(PROPOSITIONAL_AXIOMATIC_SYSTEM_SCHEMAS, formula, lines)
    # Task 9.11b

def prove_tautology(tautology: Formula) -> Proof:
    """Proves the given predicate-logic tautology.

    Parameters:
        tautology: predicate-logic tautology, whose propositional skeleton
            contains no constants or operators beyond ``'->'`` and ``'~'``, to
            prove.

    Returns:
        A valid proof of the given predicate-logic tautology from the axioms
        `PROPOSITIONAL_AXIOMATIC_SYSTEM_SCHEMAS` via only assumption lines
        and MP lines.
    """
    skeleton, substitution_map = tautology.propositional_skeleton()
    assert is_propositional_tautology(skeleton)
    assert skeleton.operators().issubset({'T', 'F', '->', '~', '&', '|'})
    skeleton_proof = prove_propositional_tautology(skeleton).clean()
    return _prove_from_skeleton_proof(tautology, skeleton_proof, substitution_map)
    # Task 9.12

def is_tautology(formula: Union[Formula, str]) -> bool:
    if isinstance(formula, str):
        formula = Formula.parse(formula)
    return is_propositional_tautology(formula.propositional_skeleton()[0])


# consider changing instantiate so that it can accept either strings or Term items as values of dict (not just strings for variable values and Term objects for constant values)

# pg. 148
# problematic example 1
# schema = Schema(Formula.parse('(Ex[R(c)]->R(c))'), {'c'})
# schema.instantiate({'c':Term('x')})
# returns unfortunate instance (Ex[R(x)]->R(x))

# problematic example 2
# schema = Schema(Formula.parse('Ey[~y=x]'), {'y'})
# schema.instantiate({'y':'x'})
# returns unfortunate instance Ex[~x=x]

# Benardete's Paradox:
# reflexive = Formula.parse('Ax[LE(x,x)]')
# transitive = Formula.parse('Ax[Ay[Az[((LE(x,y)&LE(y,z))->LE(x,z))]]]')
# asymetric = Formula.parse('Ax[Ay[((LE(x,y)&LE(y,x))->x=y)]]')
# strongly_connected = Formula.parse('Ax[Ay[(LE(x,y)|LE(y,x))]]')
# beginningless = Formula.parse('Ax[Ey[LE(y,x)]]')
# anbc = Formula.parse('Ax[Ay[(P(x)<->(LE(y,x)->~P(y)))]]')
